"""Algorithms for renumbering of counted objects, currently variables and indices."""
# Copyright (C) 2008-2024 Martin Sandve Alnæs, Anders Logg, Jørgen S. Dokken and Lawrence Mitchell
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from collections import defaultdict
from itertools import count as _count

from ufl.algorithms.map_integrands import map_integrand_dags
from ufl.core.multiindex import Index
from ufl.corealg.multifunction import MultiFunction


class IndexRelabeller(MultiFunction):
    """Renumber indices to have a consistent index numbering starting from 0."""

    def __init__(self):
        """Initialize index relabeller with a zero count."""
        super().__init__()
        count = _count()
        self.index_cache = defaultdict(lambda: Index(next(count)))

    expr = MultiFunction.reuse_if_untouched

    def multi_index(self, o):
        """Apply to multi-indices."""
        return type(o)(
            tuple(self.index_cache[i] if isinstance(i, Index) else i for i in o.indices())
        )

    def zero(self, o):
        """Apply to zero."""
        fi = o.ufl_free_indices
        fid = o.ufl_index_dimensions
        new_indices = [self.index_cache[Index(i)].count() for i in fi]
        if fi == () and fid == ():
            return o
        new_fi, new_fid = zip(*sorted(zip(new_indices, fid), key=lambda x: x[0]))
        return type(o)(o.ufl_shape, tuple(new_fi), tuple(new_fid))


def renumber_indices(form):
    """Renumber indices to have a consistent index numbering starting from 0.

    This is useful to avoid multiple kernels for the same integrand,
    but with different subdomain ids.

    Args:
        form: A UFL form, integral or expression.

    Returns:
        A new form, integral or expression with renumbered indices.
    """
    reindexer = IndexRelabeller()
    return map_integrand_dags(reindexer, form)



"""Algorithm for replacing gradients in an expression."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.algorithms.map_integrands import map_integrand_dags
from ufl.classes import ReferenceValue
from ufl.corealg.multifunction import MultiFunction, memoized_handler


class FunctionPullbackApplier(MultiFunction):
    """A pull back applier."""

    def __init__(self):
        """Initalise."""
        MultiFunction.__init__(self)

    expr = MultiFunction.reuse_if_untouched

    def terminal(self, t):
        """Apply to a terminal."""
        return t

    @memoized_handler
    def form_argument(self, o):
        """Apply to a form_argument."""
        # Represent 0-derivatives of form arguments on reference
        # element
        r = ReferenceValue(o)
        space = o.ufl_function_space()
        element = o.ufl_element()

        if r.ufl_shape != element.reference_value_shape:
            raise ValueError(
                "Expecting reference space expression with shape "
                f"'{element.reference_value_shape}', got '{r.ufl_shape}'"
            )
        f = element.pullback.apply(r)
        if f.ufl_shape != space.value_shape:
            raise ValueError(
                f"Expecting pulled back expression with shape '{space.value_shape}', "
                f"got '{f.ufl_shape}'"
            )

        assert f.ufl_shape == o.ufl_shape
        return f


def apply_function_pullbacks(expr):
    """Change representation of coefficients and arguments in an expression.

    Applies Piola mappings where applicable and represents all
    form arguments in reference value.

    Args:
        expr: An Expression
    """
    return map_integrand_dags(FunctionPullbackApplier(), expr)



from ufl import Coefficient, FunctionSpace, Mesh, TestFunction, TrialFunction, dx, triangle
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)
v = TestFunction(space)
u = TrialFunction(space)
f = Coefficient(space)

a = u * v * dx
L = f * v * dx



"""Support fpr non-literal values which are constant with respect to a domain."""

# Copyright (C) 2019 Michal Habera
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.core.terminal import Terminal
from ufl.core.ufl_type import ufl_type
from ufl.domain import as_domain
from ufl.utils.counted import Counted


@ufl_type()
class Constant(Terminal, Counted):
    """Constant."""

    _ufl_noslots_ = True

    def __init__(self, domain, shape=(), count=None):
        """Initalise."""
        Terminal.__init__(self)
        Counted.__init__(self, count, Constant)

        self._ufl_domain = as_domain(domain)
        self._ufl_shape = shape

        # Repr string is build in such way, that reconstruction
        # with eval() is possible
        self._repr = "Constant({}, {}, {})".format(
            repr(self._ufl_domain), repr(self._ufl_shape), repr(self._count)
        )

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self._ufl_shape

    def ufl_domain(self):
        """Get the UFL domain."""
        return self._ufl_domain

    def ufl_domains(self):
        """Get the UFL domains."""
        return (self.ufl_domain(),)

    def is_cellwise_constant(self):
        """Return True if the function is cellwise constant."""
        return True

    def __str__(self):
        """Format as a string."""
        return f"c_{self._count}"

    def __repr__(self):
        """Representation."""
        return self._repr

    def __eq__(self, other):
        """Check equality."""
        if not isinstance(other, Constant):
            return False
        if self is other:
            return True
        return (
            self._count == other._count
            and self._ufl_domain == other._ufl_domain
            and self._ufl_shape == self._ufl_shape
        )

    def _ufl_signature_data_(self, renumbering):
        """Signature data for constant depends on renumbering."""
        return "Constant({}, {}, {})".format(
            self._ufl_domain._ufl_signature_data_(renumbering),
            repr(self._ufl_shape),
            repr(renumbering[self]),
        )


def VectorConstant(domain, count=None):
    """Vector constant."""
    domain = as_domain(domain)
    return Constant(domain, shape=(domain.geometric_dimension(),), count=count)


def TensorConstant(domain, count=None):
    """Tensor constant."""
    domain = as_domain(domain)
    return Constant(
        domain, shape=(domain.geometric_dimension(), domain.geometric_dimension()), count=count
    )



from ufl import (
    Identity,
    Mesh,
    SpatialCoordinate,
    as_matrix,
    as_ufl,
    as_vector,
    elem_div,
    elem_mult,
    elem_op,
    sin,
    tetrahedron,
    triangle,
)
from ufl.classes import ComplexValue, Division, FloatValue, IntValue
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_scalar_casting(self):
    f = as_ufl(2.0)
    r = as_ufl(4)
    c = as_ufl(1 + 2j)
    self.assertIsInstance(f, FloatValue)
    self.assertIsInstance(r, IntValue)
    self.assertIsInstance(c, ComplexValue)
    assert float(f) == 2.0
    assert int(r) == 4
    assert complex(c) == 1.0 + 2.0j


def test_ufl_float_division(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    d = SpatialCoordinate(domain)[0] / 10.0  # TODO: Use mock instead of x
    self.assertIsInstance(d, Division)


def test_float_ufl_division(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    d = 3.14 / SpatialCoordinate(domain)[0]  # TODO: Use mock instead of x
    self.assertIsInstance(d, Division)


def test_float_division(self):
    d = as_ufl(20.0) / 10.0
    self.assertIsInstance(d, FloatValue)
    assert float(d) == 2.0


def test_int_division(self):
    # UFL treats all divisions as true division
    d = as_ufl(40) / 7
    self.assertIsInstance(d, FloatValue)
    assert float(d) == 40.0 / 7.0
    # self.assertAlmostEqual(float(d), 40 / 7.0, 15)


def test_float_int_division(self):
    d = as_ufl(20.0) / 5
    self.assertIsInstance(d, FloatValue)
    assert float(d) == 4.0


def test_floor_division_fails(self):
    f = as_ufl(2.0)
    r = as_ufl(4)
    s = as_ufl(5)
    self.assertRaises(NotImplementedError, lambda: r // 4)
    self.assertRaises(NotImplementedError, lambda: r // s)
    self.assertRaises(NotImplementedError, lambda: f // s)


def test_elem_mult(self):
    self.assertEqual(int(elem_mult(2, 3)), 6)

    v = as_vector((1, 2, 3))
    u = as_vector((4, 5, 6))
    self.assertEqual(elem_mult(v, u), as_vector((4, 10, 18)))


def test_elem_mult_on_matrices(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))

    A = as_matrix(((1, 2), (3, 4)))
    B = as_matrix(((4, 5), (6, 7)))
    self.assertEqual(elem_mult(A, B), as_matrix(((4, 10), (18, 28))))

    x, y = SpatialCoordinate(domain)
    A = as_matrix(((x, y), (3, 4)))
    B = as_matrix(((4, 5), (y, x)))
    self.assertEqual(elem_mult(A, B), as_matrix(((4 * x, 5 * y), (3 * y, 4 * x))))

    x, y = SpatialCoordinate(domain)
    A = as_matrix(((x, y), (3, 4)))
    B = Identity(2)
    self.assertEqual(elem_mult(A, B), as_matrix(((x, 0), (0, 4))))


def test_elem_div(self):
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    x, y, z = SpatialCoordinate(domain)
    A = as_matrix(((x, y, z), (3, 4, 5)))
    B = as_matrix(((7, 8, 9), (z, x, y)))
    self.assertEqual(elem_div(A, B), as_matrix(((x / 7, y / 8, z / 9), (3 / z, 4 / x, 5 / y))))


def test_elem_op(self):
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    x, y, z = SpatialCoordinate(domain)
    A = as_matrix(((x, y, z), (3, 4, 5)))
    self.assertEqual(
        elem_op(sin, A), as_matrix(((sin(x), sin(y), sin(z)), (sin(3), sin(4), sin(5))))
    )
    self.assertEqual(elem_op(sin, A).dx(0).ufl_shape, (2, 3))



__authors__ = "Martin Sandve Alnæs"
__date__ = "2008-09-06 -- 2009-02-10"

import ufl
from ufl import *  # noqa: F403
from ufl import (
    And,
    Argument,
    CellDiameter,
    CellVolume,
    Circumradius,
    Coefficient,
    Constant,
    FacetArea,
    FacetNormal,
    FunctionSpace,
    Identity,
    Jacobian,
    JacobianDeterminant,
    JacobianInverse,
    MaxFacetEdgeLength,
    Mesh,
    MinFacetEdgeLength,
    Not,
    Or,
    PermutationSymbol,
    SpatialCoordinate,
    TensorConstant,
    VectorConstant,
    acos,
    action,
    as_matrix,
    as_tensor,
    as_ufl,
    as_vector,
    asin,
    atan,
    cell_avg,
    cofac,
    conditional,
    cos,
    cosh,
    cross,
    curl,
    derivative,
    det,
    dev,
    diff,
    div,
    dot,
    dS,
    ds,
    dx,
    eq,
    exp,
    facet_avg,
    ge,
    grad,
    gt,
    i,
    inner,
    inv,
    j,
    k,
    l,
    le,
    ln,
    lt,
    nabla_div,
    nabla_grad,
    ne,
    outer,
    rot,
    sin,
    sinh,
    skew,
    sqrt,
    sym,
    tan,
    tanh,
    tetrahedron,
    tr,
    transpose,
    triangle,
    variable,
)
from ufl.algorithms import *  # noqa: F403
from ufl.classes import *  # noqa: F403
from ufl.classes import (
    Acos,
    Asin,
    Atan,
    CellCoordinate,
    Cos,
    Cosh,
    Exp,
    Expr,
    FacetJacobian,
    FacetJacobianDeterminant,
    FacetJacobianInverse,
    FloatValue,
    IntValue,
    Ln,
    Outer,
    Sin,
    Sinh,
    Sqrt,
    Tan,
    Tanh,
    all_ufl_classes,
)
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

has_repr = set()
has_dict = set()


def _test_object(a, shape, free_indices):
    # Check if instances of this type has certain memory consuming members
    if hasattr(a, "_repr"):
        has_repr.add(a.__class__.__name__)
    if hasattr(a, "__dict__"):
        has_dict.add(a.__class__.__name__)

    # Test reproduction via repr string
    r = repr(a)
    e = eval(r, globals())
    assert hash(a) == hash(e)

    # Can't really test str more than that it exists
    str(a)

    # Check that some properties are at least available
    fi = a.ufl_free_indices
    sh = a.ufl_shape

    # Compare with provided properties
    if free_indices is not None:
        free_indices = [i.count() for i in free_indices]
        if len(set(fi) ^ set(free_indices)) != 0:
            print(type(a))
            print(a)
            print(fi)
            print(free_indices)
        assert len(set(fi) ^ set(free_indices)) == 0
    if shape is not None:
        if sh != shape:
            print(("sh:", sh))
            print(("shape:", shape))
        assert sh == shape


def _test_object2(a):
    # Check if instances of this type has certain memory consuming members
    if hasattr(a, "_repr"):
        has_repr.add(a.__class__.__name__)
    if hasattr(a, "__dict__"):
        has_dict.add(a.__class__.__name__)

    # Test reproduction via repr string
    r = repr(a)
    e = eval(r, globals())
    assert hash(a) == hash(e)

    # Can't really test str more than that it exists
    str(a)


def _test_form(a):
    # Test reproduction via repr string
    r = repr(a)
    e = eval(r, globals())
    assert hash(a) == hash(e)

    # Can't really test str more than that it exists
    str(a)


def testExports(self):
    "Verify that ufl.classes exports all Expr subclasses."
    all_expr_classes = []
    for m in list(vars(ufl).values()):
        if isinstance(m, type(ufl)):
            for c in list(vars(m).values()):
                if isinstance(c, type) and issubclass(c, Expr):
                    all_expr_classes.append(c)
    missing_classes = set(c.__name__ for c in all_expr_classes) - set(
        c.__name__ for c in all_ufl_classes
    )
    if missing_classes:
        print("The following subclasses of Expr were not exported from ufl.classes:")
        print(("\n".join(sorted(missing_classes))))
    assert missing_classes == set()


def testAll(self):
    Expr.ufl_enable_profiling()

    # --- Elements:
    cell = triangle
    dim = 2

    e0 = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    e1 = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
    e2 = FiniteElement("Lagrange", cell, 1, (2, 2), identity_pullback, H1)
    e3 = MixedElement([e0, e1, e2])

    e13D = FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", cell, 1, (dim,), identity_pullback, H1))
    domain3D = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    e0_space = FunctionSpace(domain, e0)
    e1_space = FunctionSpace(domain, e1)
    e2_space = FunctionSpace(domain, e2)
    e3_space = FunctionSpace(domain, e3)
    e13d_space = FunctionSpace(domain3D, e13D)

    # --- Terminals:

    v13D = Argument(e13d_space, 3)
    f13D = Coefficient(e13d_space)

    v0 = Argument(e0_space, 4)
    v1 = Argument(e1_space, 5)
    v2 = Argument(e2_space, 6)
    v3 = Argument(e3_space, 7)

    _test_object(v0, (), ())
    _test_object(v1, (dim,), ())
    _test_object(v2, (dim, dim), ())
    _test_object(v3, (1 + dim + dim**2,), ())

    f0 = Coefficient(e0_space)
    f1 = Coefficient(e1_space)
    f2 = Coefficient(e2_space)
    f3 = Coefficient(e3_space)

    _test_object(f0, (), ())
    _test_object(f1, (dim,), ())
    _test_object(f2, (dim, dim), ())
    _test_object(f3, (1 + dim + dim**2,), ())

    c = Constant(domain)
    _test_object(c, (), ())

    c = VectorConstant(domain)
    _test_object(c, (dim,), ())

    c = TensorConstant(domain)
    _test_object(c, (dim, dim), ())

    a = FloatValue(1.23)
    _test_object(a, (), ())

    a = IntValue(123)
    _test_object(a, (), ())

    ident = Identity(1)
    _test_object(ident, (1, 1), ())
    ident = Identity(2)
    _test_object(ident, (2, 2), ())
    ident = Identity(3)
    _test_object(ident, (3, 3), ())

    e = PermutationSymbol(2)
    _test_object(e, (2, 2), ())
    e = PermutationSymbol(3)
    _test_object(e, (3, 3, 3), ())

    x = SpatialCoordinate(domain)
    _test_object(x, (dim,), ())
    xi = CellCoordinate(domain)
    _test_object(xi, (dim,), ())

    # g = CellBarycenter(domain)
    # _test_object(g, (dim,), ())
    # g = FacetBarycenter(domain)
    # _test_object(g, (dim,), ())

    g = Jacobian(domain)
    _test_object(g, (dim, dim), ())
    g = JacobianDeterminant(domain)
    _test_object(g, (), ())
    g = JacobianInverse(domain)
    _test_object(g, (dim, dim), ())

    g = FacetJacobian(domain)
    _test_object(g, (dim, dim - 1), ())
    g = FacetJacobianDeterminant(domain)
    _test_object(g, (), ())
    g = FacetJacobianInverse(domain)
    _test_object(g, (dim - 1, dim), ())

    g = FacetNormal(domain)
    _test_object(g, (dim,), ())
    # g = CellNormal(domain)
    # _test_object(g, (dim,), ())

    g = CellVolume(domain)
    _test_object(g, (), ())
    g = CellDiameter(domain)
    _test_object(g, (), ())
    g = Circumradius(domain)
    _test_object(g, (), ())
    # g = CellSurfaceArea(domain)
    # _test_object(g, (), ())

    g = FacetArea(domain)
    _test_object(g, (), ())
    g = MinFacetEdgeLength(domain)
    _test_object(g, (), ())
    g = MaxFacetEdgeLength(domain)
    _test_object(g, (), ())
    # g = FacetDiameter(domain)
    # _test_object(g, (), ())

    a = variable(v0)
    _test_object(a, (), ())
    a = variable(v1)
    _test_object(a, (dim,), ())
    a = variable(v2)
    _test_object(a, (dim, dim), ())
    a = variable(v3)
    _test_object(a, (1 + dim + dim**2,), ())
    a = variable(f0)
    _test_object(a, (), ())
    a = variable(f1)
    _test_object(a, (dim,), ())
    a = variable(f2)
    _test_object(a, (dim, dim), ())
    a = variable(f3)
    _test_object(a, (1 + dim + dim**2,), ())

    # a = MultiIndex()

    # --- Non-terminals:

    # a = Indexed()
    a = v2[i, j]
    _test_object(a, (), (i, j))
    a = v2[0, k]
    _test_object(a, (), (k,))
    a = v2[l, 1]
    _test_object(a, (), (l,))
    a = f2[i, j]
    _test_object(a, (), (i, j))
    a = f2[0, k]
    _test_object(a, (), (k,))
    a = f2[l, 1]
    _test_object(a, (), (l,))

    ident = Identity(dim)
    a = inv(ident)
    _test_object(a, (dim, dim), ())
    a = inv(v2)
    _test_object(a, (dim, dim), ())
    a = inv(f2)
    _test_object(a, (dim, dim), ())

    for v in (v0, v1, v2, v3):
        for f in (f0, f1, f2, f3):
            a = outer(v, f)
            _test_object(a, None, None)

    for v, f in zip((v0, v1, v2, v3), (f0, f1, f2, f3)):
        a = inner(v, f)
        _test_object(a, None, None)

    for v, f in zip((v1, v2, v3), (f1, f2, f3)):
        a = dot(v, f)
        sh = v.ufl_shape[:-1] + f.ufl_shape[1:]
        _test_object(a, sh, None)

    a = cross(v13D, f13D)
    _test_object(a, (3,), ())

    # a = Sum()
    a = v0 + f0 + v0
    _test_object(a, (), ())
    a = v1 + f1 + v1
    _test_object(a, (dim,), ())
    a = v2 + f2 + v2
    _test_object(a, (dim, dim), ())
    # a = Product()
    a = 3 * v0 * (2.0 * v0) * f0 * (v0 * 3.0)
    _test_object(a, (), ())
    # a = Division()
    a = v0 / 2.0
    _test_object(a, (), ())
    a = v0 / f0
    _test_object(a, (), ())
    a = v0 / (f0 + 7)
    _test_object(a, (), ())
    # a = Power()
    a = f0**3
    _test_object(a, (), ())
    a = (f0 * 2) ** 1.23
    _test_object(a, (), ())

    # a = ListTensor()
    a = as_vector([1.0, 2.0 * f0, f0**2])
    _test_object(a, (3,), ())
    a = as_matrix([[1.0, 2.0 * f0, f0**2], [1.0, 2.0 * f0, f0**2]])
    _test_object(a, (2, 3), ())
    a = as_tensor(
        [[[0.00, 0.01, 0.02], [0.10, 0.11, 0.12]], [[1.00, 1.01, 1.02], [1.10, 1.11, 1.12]]]
    )
    _test_object(a, (2, 2, 3), ())

    # a = ComponentTensor()
    a = as_vector(v1[i] * f1[j], i)
    _test_object(a, (dim,), (j,))
    a = as_matrix(v1[i] * f1[j], (j, i))
    _test_object(a, (dim, dim), ())
    a = as_tensor(v1[i] * f1[j], (i, j))
    _test_object(a, (dim, dim), ())
    a = as_tensor(v2[i, j] * f2[j, k], (i, k))
    _test_object(a, (dim, dim), ())

    a = dev(v2)
    _test_object(a, (dim, dim), ())
    a = dev(f2)
    _test_object(a, (dim, dim), ())
    a = dev(f2 * f0 + v2 * 3)
    _test_object(a, (dim, dim), ())

    a = sym(v2)
    _test_object(a, (dim, dim), ())
    a = sym(f2)
    _test_object(a, (dim, dim), ())
    a = sym(f2 * f0 + v2 * 3)
    _test_object(a, (dim, dim), ())

    a = skew(v2)
    _test_object(a, (dim, dim), ())
    a = skew(f2)
    _test_object(a, (dim, dim), ())
    a = skew(f2 * f0 + v2 * 3)
    _test_object(a, (dim, dim), ())

    a = v2.T
    _test_object(a, (dim, dim), ())
    a = f2.T
    _test_object(a, (dim, dim), ())
    a = transpose(f2 * f0 + v2 * 3)
    _test_object(a, (dim, dim), ())

    a = det(v2)
    _test_object(a, (), ())
    a = det(f2)
    _test_object(a, (), ())
    a = det(f2 * f0 + v2 * 3)
    _test_object(a, (), ())

    a = tr(v2)
    _test_object(a, (), ())
    a = tr(f2)
    _test_object(a, (), ())
    a = tr(f2 * f0 + v2 * 3)
    _test_object(a, (), ())

    a = cofac(v2)
    _test_object(a, (dim, dim), ())
    a = cofac(f2)
    _test_object(a, (dim, dim), ())
    a = cofac(f2 * f0 + v2 * 3)
    _test_object(a, (dim, dim), ())

    cond1 = le(f0, 1.0)
    cond2 = eq(3.0, f0)
    cond3 = ne(sin(f0), cos(f0))
    cond4 = lt(sin(f0), cos(f0))
    cond5 = ge(sin(f0), cos(f0))
    cond6 = gt(sin(f0), cos(f0))
    cond7 = And(cond1, cond2)
    cond8 = Or(cond1, cond2)
    cond9 = Not(cond8)
    a = conditional(cond1, 1, 2)
    b = conditional(cond2, f0**3, ln(f0))

    _test_object2(cond1)
    _test_object2(cond2)
    _test_object2(cond3)
    _test_object2(cond4)
    _test_object2(cond5)
    _test_object2(cond6)
    _test_object2(cond7)
    _test_object2(cond8)
    _test_object2(cond9)
    _test_object(a, (), ())
    _test_object(b, (), ())

    a = abs(f0)
    _test_object(a, (), ())
    a = sqrt(f0)
    _test_object(a, (), ())
    a = cos(f0)
    _test_object(a, (), ())
    a = sin(f0)
    _test_object(a, (), ())
    a = tan(f0)
    _test_object(a, (), ())
    a = cosh(f0)
    _test_object(a, (), ())
    a = sinh(f0)
    _test_object(a, (), ())
    a = tanh(f0)
    _test_object(a, (), ())
    a = exp(f0)
    _test_object(a, (), ())
    a = ln(f0)
    _test_object(a, (), ())
    a = asin(f0)
    _test_object(a, (), ())
    a = acos(f0)
    _test_object(a, (), ())
    a = atan(f0)
    _test_object(a, (), ())

    one = as_ufl(1)
    a = abs(one)
    _test_object(a, (), ())
    a = Sqrt(one)
    _test_object(a, (), ())
    a = Cos(one)
    _test_object(a, (), ())
    a = Sin(one)
    _test_object(a, (), ())
    a = Tan(one)
    _test_object(a, (), ())
    a = Cosh(one)
    _test_object(a, (), ())
    a = Sinh(one)
    _test_object(a, (), ())
    a = Tanh(one)
    _test_object(a, (), ())
    a = Acos(one)
    _test_object(a, (), ())
    a = Asin(one)
    _test_object(a, (), ())
    a = Atan(one)
    _test_object(a, (), ())
    a = Exp(one)
    _test_object(a, (), ())
    a = Ln(one)
    _test_object(a, (), ())

    # TODO:

    # a = SpatialDerivative()
    a = f0.dx(0)
    _test_object(a, (), ())
    a = f0.dx(i)
    _test_object(a, (), (i,))
    a = f0.dx(i, j, 1)
    _test_object(a, (), (i, j))

    s0 = variable(f0)
    s1 = variable(f1)
    s2 = variable(f2)
    f = dot(s0 * s1, s2)
    _test_object(s0, (), ())
    _test_object(s1, (dim,), ())
    _test_object(s2, (dim, dim), ())
    _test_object(f, (dim,), ())

    a = diff(f, s0)
    _test_object(a, (dim,), ())
    a = diff(f, s1)
    _test_object(
        a,
        (
            dim,
            dim,
        ),
        (),
    )
    a = diff(f, s2)
    _test_object(a, (dim, dim, dim), ())

    a = div(v1)
    _test_object(a, (), ())
    a = div(f1)
    _test_object(a, (), ())
    a = div(v2)
    _test_object(a, (dim,), ())
    a = div(f2)
    _test_object(a, (dim,), ())
    a = div(Outer(f1, f1))
    _test_object(a, (dim,), ())

    a = grad(v0)
    _test_object(a, (dim,), ())
    a = grad(f0)
    _test_object(a, (dim,), ())
    a = grad(v1)
    _test_object(a, (dim, dim), ())
    a = grad(f1)
    _test_object(a, (dim, dim), ())
    a = grad(f0 * v0)
    _test_object(a, (dim,), ())
    a = grad(f0 * v1)
    _test_object(a, (dim, dim), ())

    a = nabla_div(v1)
    _test_object(a, (), ())
    a = nabla_div(f1)
    _test_object(a, (), ())
    a = nabla_div(v2)
    _test_object(a, (dim,), ())
    a = nabla_div(f2)
    _test_object(a, (dim,), ())
    a = nabla_div(Outer(f1, f1))
    _test_object(a, (dim,), ())

    a = nabla_grad(v0)
    _test_object(a, (dim,), ())
    a = nabla_grad(f0)
    _test_object(a, (dim,), ())
    a = nabla_grad(v1)
    _test_object(a, (dim, dim), ())
    a = nabla_grad(f1)
    _test_object(a, (dim, dim), ())
    a = nabla_grad(f0 * v0)
    _test_object(a, (dim,), ())
    a = nabla_grad(f0 * v1)
    _test_object(a, (dim, dim), ())

    a = curl(v13D)
    _test_object(a, (3,), ())
    a = curl(f13D)
    _test_object(a, (3,), ())
    a = rot(v1)
    _test_object(a, (), ())
    a = rot(f1)
    _test_object(a, (), ())

    # a = PositiveRestricted(v0)
    # _test_object(a, (), ())
    a = v0("+")
    _test_object(a, (), ())
    a = v0("+") * f0
    _test_object(a, (), ())

    # a = NegativeRestricted(v0)
    # _test_object(a, (), ())
    a = v0("-")
    _test_object(a, (), ())
    a = v0("-") + f0
    _test_object(a, (), ())

    a = cell_avg(v0)
    _test_object(a, (), ())
    a = facet_avg(v0)
    _test_object(a, (), ())
    a = cell_avg(v1)
    _test_object(a, (dim,), ())
    a = facet_avg(v1)
    _test_object(a, (dim,), ())
    a = cell_avg(v1)[i]
    _test_object(a, (), (i,))
    a = facet_avg(v1)[i]
    _test_object(a, (), (i,))

    # --- Integrals:

    a = v0 * dx
    _test_form(a)
    a = v0 * dx(0)
    _test_form(a)
    a = v0 * dx(1)
    _test_form(a)
    a = v0 * ds
    _test_form(a)
    a = v0 * ds(0)
    _test_form(a)
    a = v0 * ds(1)
    _test_form(a)
    a = v0 * dS
    _test_form(a)
    a = v0 * dS(0)
    _test_form(a)
    a = v0 * dS(1)
    _test_form(a)

    a = v0 * dot(v1, f1) * dx
    _test_form(a)
    a = v0 * dot(v1, f1) * dx(0)
    _test_form(a)
    a = v0 * dot(v1, f1) * dx(1)
    _test_form(a)
    a = v0 * dot(v1, f1) * ds
    _test_form(a)
    a = v0 * dot(v1, f1) * ds(0)
    _test_form(a)
    a = v0 * dot(v1, f1) * ds(1)
    _test_form(a)
    a = v0 * dot(v1, f1) * dS
    _test_form(a)
    a = v0 * dot(v1, f1) * dS(0)
    _test_form(a)
    a = v0 * dot(v1, f1) * dS(1)
    _test_form(a)

    # --- Form transformations:

    a = f0 * v0 * dx + f0 * v0 * dot(f1, v1) * dx
    # b = lhs(a) # TODO
    # c = rhs(a) # TODO
    d = derivative(a, f1, v1)
    f = action(d)
    # e = action(b)

    # --- Check which classes have been created
    ic, dc = Expr.ufl_disable_profiling()

    constructed = set()
    unused = set(Expr._ufl_all_classes_)
    for cls in Expr._ufl_all_classes_:
        tc = cls._ufl_typecode_
        if ic[tc]:
            constructed.add(cls)
        if cls._ufl_is_abstract_:
            unused.remove(cls)

    if unused:
        print()
        print("The following classes were never instantiated in class coverage test:")
        print(("\n".join(sorted(map(str, unused)))))
        print()

    # --- Check which classes had certain member variables
    if has_repr:
        print()
        print("The following classes contain a _repr member:")
        print(("\n".join(sorted(map(str, has_repr)))))
        print()

    if has_dict:
        print()
        print("The following classes contain a __dict__ member:")
        print(("\n".join(sorted(map(str, has_dict)))))
        print()

    # TODO: Add tests for bessel functions:
    #   BesselI
    #   BesselJ
    #   BesselK
    #   BesselY
    #   Erf
    # TODO: Add tests for:
    #   Label



import pytest

import ufl.algorithms
import ufl.classes
from ufl import (
    Argument,
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    as_matrix,
    as_tensor,
    as_vector,
    cos,
    dx,
    exp,
    i,
    indices,
    interval,
    j,
    k,
    l,
    outer,
    sin,
    triangle,
)
from ufl.classes import IndexSum
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

# TODO: add more expressions to test as many possible combinations of
# index notation as feasible...


def test_vector_indices(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    u = Argument(space, 2)
    f = Coefficient(space)
    u[i] * f[i] * dx
    u[j] * f[j] * dx


def test_tensor_indices(self):
    element = FiniteElement("Lagrange", triangle, 1, (2, 2), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    u = Argument(space, 2)
    f = Coefficient(space)
    u[i, j] * f[i, j] * dx
    u[j, i] * f[i, j] * dx
    u[j, i] * f[j, i] * dx
    with pytest.raises(BaseException):
        (u[i, i] + f[j, i]) * dx


def test_indexed_sum1(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    u = Argument(space, 2)
    f = Coefficient(space)
    a = u[i] + f[i]
    with pytest.raises(BaseException):
        a * dx


def test_indexed_sum2(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = Argument(space, 2)
    u = Argument(space, 3)
    f = Coefficient(space)
    a = u[j] + f[j] + v[j] + 2 * v[j] + exp(u[i] * u[i]) / 2 * f[j]
    with pytest.raises(BaseException):
        a * dx


def test_indexed_sum3(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    u = Argument(space, 2)
    f = Coefficient(space)
    with pytest.raises(BaseException):
        u[i] + f[j]


def test_indexed_function1(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = Argument(space, 2)
    u = Argument(space, 3)
    f = Coefficient(space)
    aarg = (u[i] + f[i]) * v[i]
    exp(aarg) * dx


def test_indexed_function2(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = Argument(space, 2)
    u = Argument(space, 3)
    f = Coefficient(space)
    bfun = cos(f[0])
    left = u[i] + f[i]
    right = v[i] * bfun
    assert len(left.ufl_free_indices) == 1
    assert left.ufl_free_indices[0] == i.count()
    assert len(right.ufl_free_indices) == 1
    assert right.ufl_free_indices[0] == i.count()
    left * right * dx


def test_indexed_function3(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    Argument(space, 2)
    u = Argument(space, 3)
    f = Coefficient(space)
    with pytest.raises(BaseException):
        sin(u[i] + f[i]) * dx


def test_vector_from_indices(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)

    # legal
    vv = as_vector(u[i], i)
    uu = as_vector(v[j], j)
    w = v + u
    ww = vv + uu
    assert len(vv.ufl_shape) == 1
    assert len(uu.ufl_shape) == 1
    assert len(w.ufl_shape) == 1
    assert len(ww.ufl_shape) == 1


def test_matrix_from_indices(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)

    A = as_matrix(u[i] * v[j], (i, j))
    B = as_matrix(v[k] * v[k] * u[i] * v[j], (j, i))
    C = A + A
    C = B + B
    D = A + B
    assert len(A.ufl_shape) == 2
    assert len(B.ufl_shape) == 2
    assert len(C.ufl_shape) == 2
    assert len(D.ufl_shape) == 2


def test_vector_from_list(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)

    # create vector from list
    vv = as_vector([u[0], v[0]])
    ww = vv + vv
    assert len(vv.ufl_shape) == 1
    assert len(ww.ufl_shape) == 1


def test_matrix_from_list(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)

    # create matrix from list
    A = as_matrix([[u[0], u[1]], [v[0], v[1]]])
    # create matrix from indices
    B = as_matrix((v[k] * v[k]) * u[i] * v[j], (j, i))
    # Test addition
    C = A + A
    C = B + B
    D = A + B
    assert len(A.ufl_shape) == 2
    assert len(B.ufl_shape) == 2
    assert len(C.ufl_shape) == 2
    assert len(D.ufl_shape) == 2


def test_tensor(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)
    g = Coefficient(space)

    # define the components of a fourth order tensor
    Cijkl = u[i] * v[j] * f[k] * g[l]
    assert len(Cijkl.ufl_shape) == 0
    assert set(Cijkl.ufl_free_indices) == {i.count(), j.count(), k.count(), l.count()}

    # make it a tensor
    C = as_tensor(Cijkl, (i, j, k, l))
    assert len(C.ufl_shape) == 4
    self.assertSameIndices(C, ())

    # get sub-matrix
    A = C[:, :, 0, 0]
    assert len(A.ufl_shape) == 2
    self.assertSameIndices(A, ())
    A = C[:, :, i, j]
    assert len(A.ufl_shape) == 2
    assert set(A.ufl_free_indices) == {i.count(), j.count()}

    # legal?
    vv = as_vector([u[i], v[i]])
    f[i] * vv  # this is well defined: ww = sum_i <f_i*u_i, f_i*v_i>

    # illegal
    with pytest.raises(BaseException):
        as_vector([u[i], v[j]])

    # illegal
    with pytest.raises(BaseException):
        as_matrix([[u[0], u[1]], [v[0]]])


def test_indexed(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)
    Coefficient(space)
    i, j, k, l = indices(4)  # noqa: E741

    a = v[i]
    self.assertSameIndices(a, (i,))

    a = outer(v, u)[i, j]
    self.assertSameIndices(a, (i, j))

    a = outer(v, u)[i, i]
    self.assertSameIndices(a, ())
    self.assertIsInstance(a, IndexSum)


def test_spatial_derivative(self):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)
    i, j, k, l = indices(4)  # noqa: E741
    d = 2

    a = v[i].dx(i)
    self.assertSameIndices(a, ())
    self.assertIsInstance(a, IndexSum)
    assert a.ufl_shape == ()

    a = v[i].dx(j)
    self.assertSameIndices(a, (i, j))
    self.assertNotIsInstance(a, IndexSum)
    assert a.ufl_shape == ()

    a = (v[i] * u[j]).dx(i, j)
    self.assertSameIndices(a, ())
    self.assertIsInstance(a, IndexSum)
    assert a.ufl_shape == ()

    a = v.dx(i, j)
    # self.assertSameIndices(a, (i,j))
    assert set(a.ufl_free_indices) == {j.count(), i.count()}
    self.assertNotIsInstance(a, IndexSum)
    assert a.ufl_shape == (d,)

    a = v[i].dx(0)
    self.assertSameIndices(a, (i,))
    self.assertNotIsInstance(a, IndexSum)
    assert a.ufl_shape == ()

    a = (v[i] * u[j]).dx(0, 1)
    assert set(a.ufl_free_indices) == {i.count(), j.count()}
    self.assertNotIsInstance(a, IndexSum)
    assert a.ufl_shape == ()

    a = v.dx(i)[i]
    self.assertSameIndices(a, ())
    self.assertIsInstance(a, IndexSum)
    assert a.ufl_shape == ()


def test_renumbering(self):
    """Test that kernels with common integral data, but different index numbering,
    are correctly renumbered."""
    cell = interval
    mesh = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    V = FunctionSpace(mesh, FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    v = TestFunction(V)
    u = TrialFunction(V)
    i = indices(1)
    a0 = u[i].dx(0) * v[i].dx(0) * ufl.dx((1))
    a1 = (
        u[i].dx(0)
        * v[i].dx(0)
        * ufl.dx(
            (
                2,
                3,
            )
        )
    )
    form_data = ufl.algorithms.compute_form_data(
        a0 + a1,
        do_apply_function_pullbacks=True,
        do_apply_integral_scaling=True,
        do_apply_geometry_lowering=True,
        preserve_geometry_types=(ufl.classes.Jacobian,),
        do_apply_restrictions=True,
        do_append_everywhere_integrals=False,
    )

    assert len(form_data.integral_data) == 1



"""Algorithm for splitting a Coefficient or Argument into subfunctions."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008

from ufl.domain import extract_unique_domain
from ufl.functionspace import FunctionSpace
from ufl.indexed import Indexed
from ufl.permutation import compute_indices
from ufl.tensors import ListTensor, as_matrix, as_vector
from ufl.utils.indexflattening import flatten_multiindex, shape_to_strides
from ufl.utils.sequences import product


def split(v):
    """Split a coefficient or argument.

    If v is a Coefficient or Argument in a mixed space, returns a tuple
    with the function components corresponding to the subelements.
    """
    domain = extract_unique_domain(v)

    # Default range is all of v
    begin = 0
    end = None

    if isinstance(v, Indexed):
        # Special case: split previous output of split again
        # Consistent with simple element, just return function in a tuple
        return (v,)

    elif isinstance(v, ListTensor):
        # Special case: split previous output of split again
        ops = v.ufl_operands
        if all(isinstance(comp, Indexed) for comp in ops):
            args = [comp.ufl_operands[0] for comp in ops]
            if all(args[0] == args[i] for i in range(1, len(args))):
                # Get innermost terminal here and its element
                v = args[0]
                # Get relevant range of v components
                (begin,) = ops[0].ufl_operands[1]
                (end,) = ops[-1].ufl_operands[1]
                begin = int(begin)
                end = int(end) + 1
            else:
                raise ValueError(f"Don't know how to split {v}.")
        else:
            raise ValueError(f"Don't know how to split {v}.")

    # Special case: simple element, just return function in a tuple
    element = v.ufl_element()
    if element.num_sub_elements == 0:
        assert end is None
        return (v,)

    if len(v.ufl_shape) != 1:
        raise ValueError(
            "Don't know how to split tensor valued mixed functions without flattened index space."
        )

    # Compute value size and set default range end
    value_size = v.ufl_function_space().value_size
    if end is None:
        end = value_size
    else:
        # Recursively dive into mixedelement in to subelement
        # corresponding to beginning of range
        j = begin
        while True:
            for e in element.sub_elements:
                if j < FunctionSpace(domain, e).value_size:
                    element = e
                    break
                j -= FunctionSpace(domain, e).value_size
            # Then break when we find the subelement that covers the whole range
            if FunctionSpace(domain, element).value_size == (end - begin):
                break

    # Build expressions representing the subfunction of v for each subelement
    offset = begin
    sub_functions = []
    for i, e in enumerate(element.sub_elements):
        # Get shape, size, indices, and v components
        # corresponding to subelement value
        shape = FunctionSpace(domain, e).value_shape
        strides = shape_to_strides(shape)
        rank = len(shape)
        sub_size = product(shape)
        subindices = [flatten_multiindex(c, strides) for c in compute_indices(shape)]
        components = [v[k + offset] for k in subindices]

        # Shape components into same shape as subelement
        if rank == 0:
            (subv,) = components
        elif rank <= 1:
            subv = as_vector(components)
        elif rank == 2:
            subv = as_matrix(
                [components[i * shape[1] : (i + 1) * shape[1]] for i in range(shape[0])]
            )
        else:
            raise ValueError(
                f"Don't know how to split functions with sub functions of rank {rank}."
            )

        offset += sub_size
        sub_functions.append(subv)

    if end != offset:
        raise ValueError(
            "Function splitting failed to extract components for whole intended range."
        )

    return tuple(sub_functions)



"""Operators.

This module extends the form language with free function operators,
which are either already available as member functions on UFL objects
or defined as compound operators involving basic operations on the UFL
objects.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs and Anders Logg
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Kristian B. Oelgaard, 2011
# Modified by Massimiliano Leoni, 2016.

import operator
import warnings

from ufl import sobolevspace
from ufl.algebra import Conj, Imag, Real
from ufl.averaging import CellAvg, FacetAvg
from ufl.checks import is_cellwise_constant
from ufl.coefficient import Coefficient
from ufl.conditional import (
    EQ,
    NE,
    AndCondition,
    Conditional,
    MaxValue,
    MinValue,
    NotCondition,
    OrCondition,
)
from ufl.constantvalue import ComplexValue, RealValue, Zero, as_ufl
from ufl.differentiation import Curl, Div, Grad, NablaDiv, NablaGrad, VariableDerivative
from ufl.domain import extract_domains
from ufl.form import Form
from ufl.geometry import FacetNormal, SpatialCoordinate
from ufl.indexed import Indexed
from ufl.mathfunctions import (
    Acos,
    Asin,
    Atan,
    Atan2,
    BesselI,
    BesselJ,
    BesselK,
    BesselY,
    Cos,
    Cosh,
    Erf,
    Exp,
    Ln,
    Sin,
    Sinh,
    Sqrt,
    Tan,
    Tanh,
)
from ufl.tensoralgebra import (
    Cofactor,
    Cross,
    Determinant,
    Deviatoric,
    Dot,
    Inner,
    Inverse,
    Outer,
    Perp,
    Skew,
    Sym,
    Trace,
    Transposed,
)
from ufl.tensors import ListTensor, as_matrix, as_tensor, as_vector
from ufl.variable import Variable

# --- Basic operators ---


def rank(f):
    """The rank of f."""
    f = as_ufl(f)
    return len(f.ufl_shape)


def shape(f):
    """The shape of f."""
    f = as_ufl(f)
    return f.ufl_shape


# --- Complex operators ---


def conj(f):
    """The complex conjugate of f."""
    f = as_ufl(f)
    return Conj(f)


# Alias because both conj and conjugate are in numpy and we wish to be
# consistent.
conjugate = conj


def real(f):
    """The real part of f."""
    f = as_ufl(f)
    return Real(f)


def imag(f):
    """The imaginary part of f."""
    f = as_ufl(f)
    return Imag(f)


# --- Elementwise tensor operators ---


def elem_op_items(op_ind, indices, *args):
    """Elem op items."""
    sh = args[0].ufl_shape
    indices = tuple(indices)
    n = sh[len(indices)]

    def extind(ii):
        return indices + (ii,)

    if len(sh) == len(indices) + 1:
        return [op_ind(extind(i), *args) for i in range(n)]
    else:
        return [elem_op_items(op_ind, extind(i), *args) for i in range(n)]


def elem_op(op, *args):
    """Apply element-wise operations.

    Take the element-wise application of operator op on scalar values
    from one or more tensor arguments.
    """
    args = [as_ufl(arg) for arg in args]
    sh = args[0].ufl_shape
    if not all(sh == x.ufl_shape for x in args):
        raise ValueError("Cannot take element-wise operation with different shapes.")

    if sh == ():
        return op(*args)

    def op_ind(ind, *args):
        return op(*[x[ind] for x in args])

    return as_tensor(elem_op_items(op_ind, (), *args))


def elem_mult(A, B):
    """Take the elementwise multiplication of tensors A and B with the same shape."""
    return elem_op(operator.mul, A, B)


def elem_div(A, B):
    """Take the elementwise division of tensors A and B with the same shape."""
    return elem_op(operator.truediv, A, B)


def elem_pow(A, B):
    """Take the elementwise power of tensors A and B with the same shape."""
    return elem_op(operator.pow, A, B)


# --- Tensor operators ---


def transpose(A):
    """Take the transposed of tensor A."""
    A = as_ufl(A)
    if A.ufl_shape == ():
        return A
    return Transposed(A)


def outer(*operands):
    """Take the outer product of two or more operands.

    The complex conjugate of the first argument is taken.
    """
    n = len(operands)
    if n == 1:
        return operands[0]
    elif n == 2:
        a, b = operands
    else:
        a = outer(*operands[:-1])
        b = operands[-1]
    a = as_ufl(a)
    b = as_ufl(b)
    if a.ufl_shape == () and b.ufl_shape == ():
        return Conj(a) * b
    return Outer(a, b)


def inner(a, b):
    """Take the inner product of a and b.

    The complex conjugate of the second argument is taken.
    """
    a = as_ufl(a)
    b = as_ufl(b)
    if a.ufl_shape == () and b.ufl_shape == ():
        return a * Conj(b)
    return Inner(a, b)


def dot(a, b):
    """Take the dot product of a and b.

    This won't take the complex conjugate of the second argument.
    """
    a = as_ufl(a)
    b = as_ufl(b)
    if a.ufl_shape == () and b.ufl_shape == ():
        return a * b
    return Dot(a, b)


def perp(v):
    """Take the perp of v.

    I.e. :math:`(-v_1, +v_0)`.
    """
    v = as_ufl(v)
    if v.ufl_shape != (2,):
        raise ValueError("Expecting a 2D vector expression.")
    return Perp(v)


def cross(a, b):
    """Take the cross product of a and b."""
    a = as_ufl(a)
    b = as_ufl(b)
    return Cross(a, b)


def det(A):
    """Take the determinant of A."""
    A = as_ufl(A)
    if A.ufl_shape == ():
        return A
    return Determinant(A)


def inv(A):
    """Take the inverse of A."""
    A = as_ufl(A)
    if A.ufl_shape == ():
        return 1 / A
    return Inverse(A)


def cofac(A):
    """Take the cofactor of A."""
    A = as_ufl(A)
    return Cofactor(A)


def tr(A):
    """Take the trace of A."""
    A = as_ufl(A)
    return Trace(A)


def diag(A):
    """Diagonal ranl-2 tensor.

    Take the diagonal part of rank 2 tensor A or make a diagonal rank 2
    tensor from a rank 1 tensor.

    Always returns a rank 2 tensor. See also diag_vector.
    """
    # TODO: Make a compound type or two for this operator

    # Get and check dimensions
    r = len(A.ufl_shape)
    if r == 1:
        (n,) = A.ufl_shape
    elif r == 2:
        m, n = A.ufl_shape
        if m != n:
            raise ValueError("Can only take diagonal of square tensors.")
    else:
        raise ValueError("Expecting rank 1 or 2 tensor.")

    # Build matrix row by row
    rows = []
    for i in range(n):
        row = [0] * n
        row[i] = A[i] if r == 1 else A[i, i]
        rows.append(row)
    return as_matrix(rows)


def diag_vector(A):
    """Take the diagonal part of rank 2 tensor A and return as a vector.

    See also diag.
    """
    # TODO: Make a compound type for this operator

    # Get and check dimensions
    if len(A.ufl_shape) != 2:
        raise ValueError("Expecting rank 2 tensor.")
    m, n = A.ufl_shape
    if m != n:
        raise ValueError("Can only take diagonal of square tensors.")

    # Return diagonal vector
    return as_vector([A[i, i] for i in range(n)])


def dev(A):
    """Take the deviatoric part of A."""
    A = as_ufl(A)
    return Deviatoric(A)


def skew(A):
    """Take the skew symmetric part of A."""
    A = as_ufl(A)
    return Skew(A)


def sym(A):
    """Take the symmetric part of A."""
    A = as_ufl(A)
    return Sym(A)


# --- Differential operators


def Dx(f, *i):
    """Take the partial derivative of f with respect to spatial variable number i.

    Equivalent to f.dx(*i).
    """
    f = as_ufl(f)
    return f.dx(*i)


def Dn(f):
    """Take the directional derivative of f in the facet normal direction.

    The facet normal is Dn(f) := dot(grad(f), n).
    """
    f = as_ufl(f)
    if is_cellwise_constant(f):
        return Zero(f.ufl_shape, f.ufl_free_indices, f.ufl_index_dimensions)
    return dot(grad(f), FacetNormal(f.ufl_domain()))


def diff(f, v):
    """Take the derivative of f with respect to the variable v.

    If f is a form, diff is applied to each integrand.
    """
    # Apply to integrands
    if isinstance(f, Form):
        from ufl.algorithms.map_integrands import map_integrands

        return map_integrands(lambda e: diff(e, v), f)

    # Apply to expression
    f = as_ufl(f)
    if isinstance(v, SpatialCoordinate):
        return grad(f)
    elif isinstance(v, (Variable, Coefficient)):
        return VariableDerivative(f, v)
    else:
        raise ValueError("Expecting a Variable or SpatialCoordinate in diff.")


def grad(f):
    """Take the gradient of f.

    This operator follows the grad convention where
      grad(s)[i] = s.dx(i)
      grad(v)[i,j] = v[i].dx(j)
      grad(T)[:,i] = T[:].dx(i)
    for scalar expressions s, vector expressions v,
    and arbitrary rank tensor expressions T.

    See also: :py:func:`nabla_grad`
    """
    f = as_ufl(f)
    return Grad(f)


def div(f):
    """Take the divergence of f.

    This operator follows the div convention where
      div(v) = v[i].dx(i)
      div(T)[:] = T[:,i].dx(i)
    for vector expressions v, and
    arbitrary rank tensor expressions T.

    See also: :py:func:`nabla_div`
    """
    f = as_ufl(f)
    return Div(f)


def nabla_grad(f):
    """Take the gradient of f.

    This operator follows the grad convention where
      nabla_grad(s)[i] = s.dx(i)
      nabla_grad(v)[i,j] = v[j].dx(i)
      nabla_grad(T)[i,:] = T[:].dx(i)
    for scalar expressions s, vector expressions v,
    and arbitrary rank tensor expressions T.

    See also: :py:func:`grad`
    """
    f = as_ufl(f)
    return NablaGrad(f)


def nabla_div(f):
    """Take the divergence of f.

    This operator follows the div convention where
      nabla_div(v) = v[i].dx(i)
      nabla_div(T)[:] = T[i,:].dx(i)
    for vector expressions v, and
    arbitrary rank tensor expressions T.

    See also: :py:func:`div`
    """
    f = as_ufl(f)
    return NablaDiv(f)


def curl(f):
    """Take the curl of f."""
    f = as_ufl(f)
    return Curl(f)


rot = curl


# --- DG operators ---


def jump(v, n=None):
    """Take the jump of v across a facet."""
    v = as_ufl(v)
    is_constant = len(extract_domains(v)) > 0
    if is_constant:
        if n is None:
            return v("+") - v("-")
        r = len(v.ufl_shape)
        if r == 0:
            return v("+") * n("+") + v("-") * n("-")
        else:
            return dot(v("+"), n("+")) + dot(v("-"), n("-"))
    else:
        warnings.warn(
            "Returning zero from jump of expression without a domain. "
            "This may be erroneous if a dolfin.Expression is involved."
        )
        # FIXME: Is this right? If v has no domain, it doesn't depend
        # on anything spatially variable or any form arguments, and
        # thus the jump is zero. In other words, I'm assuming that "v
        # has no geometric domains" is equivalent with "v is a spatial
        # constant".  Update: This is NOT true for
        # jump(Expression("x[0]")) from dolfin.
        return Zero(v.ufl_shape, v.ufl_free_indices, v.ufl_index_dimensions)


def avg(v):
    """Take the average of v across a facet."""
    v = as_ufl(v)
    return 0.5 * (v("+") + v("-"))


def cell_avg(f):
    """Take the average of v over a cell."""
    return CellAvg(f)


def facet_avg(f):
    """Take the average of v over a facet."""
    return FacetAvg(f)


# --- Other operators ---


def variable(e):
    """Define a variable representing the given expression.

    See also diff().
    """
    e = as_ufl(e)
    return Variable(e)


# --- Conditional expressions ---


def conditional(condition, true_value, false_value):
    """A conditional expression.

    This takes the value of true_value
    when condition evaluates to true and false_value otherwise.
    """
    return Conditional(condition, true_value, false_value)


def eq(left, right):
    """A boolean expression (left == right) for use with conditional."""
    return EQ(left, right)


def ne(left, right):
    """A boolean expression (left != right) for use with conditional."""
    return NE(left, right)


def le(left, right):
    """A boolean expression (left <= right) for use with conditional."""
    return as_ufl(left) <= as_ufl(right)


def ge(left, right):
    """A boolean expression (left >= right) for use with conditional."""
    return as_ufl(left) >= as_ufl(right)


def lt(left, right):
    """A boolean expression (left < right) for use with conditional."""
    return as_ufl(left) < as_ufl(right)


def gt(left, right):
    """A boolean expression (left > right) for use with conditional."""
    return as_ufl(left) > as_ufl(right)


def And(left, right):
    """A boolean expression (left and right) for use with conditional."""
    return AndCondition(left, right)


def Or(left, right):
    """A boolean expression (left or right) for use with conditional."""
    return OrCondition(left, right)


def Not(condition):
    """A boolean expression (not condition) for use with conditional."""
    return NotCondition(condition)


def sign(x):
    """Return the sign of x.

    This returns +1 if x is positive, -1 if x is negative, and 0 if x is 0.
    """
    # TODO: Add a Sign type for this?
    return conditional(eq(x, 0), 0, conditional(lt(x, 0), -1, +1))


def max_value(x, y):
    """Take the maximum of x and y."""
    x = as_ufl(x)
    y = as_ufl(y)
    return MaxValue(x, y)


def min_value(x, y):
    """Take the minimum of x and y."""
    x = as_ufl(x)
    y = as_ufl(y)
    return MinValue(x, y)


# --- Math functions ---


def _mathfunction(f, cls):
    """A mat function."""
    f = as_ufl(f)
    r = cls(f)
    if isinstance(r, (RealValue, Zero, int, float)):
        return float(r)
    if isinstance(r, (ComplexValue, complex)):
        return complex(r)
    return r


def sqrt(f):
    """Take the square root of f."""
    return _mathfunction(f, Sqrt)


def exp(f):
    """Take the exponential of f."""
    return _mathfunction(f, Exp)


def ln(f):
    """Take the natural logarithm of f."""
    return _mathfunction(f, Ln)


def cos(f):
    """Take the cosine of f."""
    return _mathfunction(f, Cos)


def sin(f):
    """Take the sine of f."""
    return _mathfunction(f, Sin)


def tan(f):
    """Take the tangent of f."""
    return _mathfunction(f, Tan)


def cosh(f):
    """Take the hyperbolic cosine of f."""
    return _mathfunction(f, Cosh)


def sinh(f):
    """Take the hyperbolic sine of f."""
    return _mathfunction(f, Sinh)


def tanh(f):
    """Take the hyperbolic tangent of f."""
    return _mathfunction(f, Tanh)


def acos(f):
    """Take the inverse cosine of f."""
    return _mathfunction(f, Acos)


def asin(f):
    """Take the inverse sine of f."""
    return _mathfunction(f, Asin)


def atan(f):
    """Take the inverse tangent of f."""
    return _mathfunction(f, Atan)


def atan2(f1, f2):
    """Take the inverse tangent with two the arguments f1 and f2."""
    f1 = as_ufl(f1)
    f2 = as_ufl(f2)
    if isinstance(f1, (ComplexValue, complex)) or isinstance(f2, (ComplexValue, complex)):
        raise TypeError("atan2 is incompatible with complex numbers.")
    r = Atan2(f1, f2)
    if isinstance(r, (RealValue, Zero, int, float)):
        return float(r)
    if isinstance(r, (ComplexValue, complex)):
        return complex(r)
    return r


def erf(f):
    """Take the error function of f."""
    return _mathfunction(f, Erf)


def bessel_J(nu, f):
    """Cylindrical Bessel function of the first kind."""
    nu = as_ufl(nu)
    f = as_ufl(f)
    return BesselJ(nu, f)


def bessel_Y(nu, f):
    """Cylindrical Bessel function of the second kind."""
    nu = as_ufl(nu)
    f = as_ufl(f)
    return BesselY(nu, f)


def bessel_I(nu, f):
    """Regular modified cylindrical Bessel function."""
    nu = as_ufl(nu)
    f = as_ufl(f)
    return BesselI(nu, f)


def bessel_K(nu, f):
    """Irregular modified cylindrical Bessel function."""
    nu = as_ufl(nu)
    f = as_ufl(f)
    return BesselK(nu, f)


# --- Special function for exterior_derivative


def exterior_derivative(f):
    """Take the exterior derivative of f.

    The exterior derivative uses the element Sobolev space to
    determine whether id, grad, curl or div should be used.

    Note that this uses the grad and div operators,
    as opposed to nabla_grad and nabla_div.
    """
    # Extract the element from the input f
    if isinstance(f, Indexed):
        expression, indices = f.ufl_operands
        if len(indices) > 1:
            raise NotImplementedError
        index = int(indices[0])
        element = expression.ufl_element()
        while index != 0:
            for e in element.sub_elements:
                if e.value_size > index:
                    element = e
                    break
                index -= e.value_size
    elif isinstance(f, ListTensor):
        f0 = f.ufl_operands[0]
        f0expr, f0indices = f0.ufl_operands  # FIXME: Assumption on type of f0!!!
        if len(f0indices) > 1:
            raise NotImplementedError
        index = int(f0indices[0])
        element = f0expr.ufl_element()
        while index != 0:
            for e in element.sub_elements:
                if e.value_size > index:
                    element = e
                    break
                index -= e.value_size
    else:
        try:
            element = f.ufl_element()
        except Exception:
            raise ValueError(f"Unable to determine element from {f}")

    domain = f.ufl_domain()

    gdim = domain.geometric_dimension()
    space = element.sobolev_space

    if space == sobolevspace.L2:
        return f
    elif space == sobolevspace.H1:
        if gdim == 1:
            return grad(f)[0]  # Special-case 1D vectors as scalars
        return grad(f)
    elif space == sobolevspace.HCurl:
        return curl(f)
    elif space == sobolevspace.HDiv:
        return div(f)
    else:
        raise ValueError(f"Unable to determine exterior_derivative for element '{element!r}'")



import pytest

import ufl
import ufl.algorithms
from ufl.finiteelement import FiniteElement, MixedElement


def epsilon(u):
    return ufl.sym(ufl.grad(u))


def sigma(u, p):
    return epsilon(u) - p * ufl.Identity(u.ufl_shape[0])


@pytest.mark.parametrize("rank", [0, 1, 2])
def test_extract_blocks(rank):
    """Test extractions of blocks from mixed function space."""
    cell = ufl.triangle
    domain = ufl.Mesh(FiniteElement("Lagrange", cell, 1, (2,), ufl.identity_pullback, ufl.H1))
    fe_scalar = FiniteElement("Lagrange", cell, 1, (), ufl.identity_pullback, ufl.H1)
    fe_vector = FiniteElement("Lagrange", cell, 1, (2,), ufl.identity_pullback, ufl.H1)

    me = MixedElement([fe_vector, fe_scalar])

    # # Function spaces
    W = ufl.FunctionSpace(domain, me)
    V = ufl.FunctionSpace(domain, fe_vector)
    Q = ufl.FunctionSpace(domain, fe_scalar)

    if rank == 0:
        wh = ufl.Coefficient(W)
        uh, ph = ufl.split(wh)
        # Test that functionals return the identity
        J = ufl.inner(sigma(uh, ph), sigma(uh, ph)) * ufl.dx
        J0 = ufl.extract_blocks(J, 0)
        assert len(J0) == 1
        assert J == J0[0]
    elif rank == 1:

        def rhs(uh, ph, v, q):
            F_0 = ufl.inner(sigma(uh, ph), epsilon(v)) * ufl.dx(domain=domain)
            F_1 = ufl.div(uh) * q * ufl.dx
            return F_0, F_1

        wh = ufl.Coefficient(W)
        uh, ph = ufl.split(wh)
        v, q = ufl.TestFunctions(W)
        F = sum(rhs(uh, ph, v, q))

        v_ = ufl.TestFunction(V)
        q_ = ufl.TestFunction(Q)
        F_sub = rhs(uh, ph, ufl.as_vector([vi for vi in v_]), q_)

        F_0_ext = ufl.extract_blocks(F, 0)
        assert F_sub[0].signature() == F_0_ext.signature()

        F_1_ext = ufl.extract_blocks(F, 1)
        assert F_sub[1].signature() == F_1_ext.signature()
    elif rank == 2:

        def lhs(u, p, v, q):
            J_00 = ufl.inner(u, v) * ufl.dx(domain=domain)
            J_01 = ufl.div(v) * p * ufl.dx
            J_10 = q * ufl.div(u) * ufl.dx
            J_11 = ufl.inner(ufl.grad(p), ufl.grad(q)) * ufl.dx
            return J_00, J_01, J_10, J_11

        v_ = ufl.TestFunction(V)
        q_ = ufl.TestFunction(Q)
        u_ = ufl.TrialFunction(V)
        p_ = ufl.TrialFunction(Q)
        J_sub = lhs(ufl.as_vector([ui for ui in u_]), p_, ufl.as_vector([vi for vi in v_]), q_)

        v, q = ufl.TestFunctions(W)
        uh, ph = ufl.TrialFunctions(W)
        J = sum(lhs(uh, ph, v, q))

        for i in range(2):
            for j in range(2):
                J_ij_ext = ufl.extract_blocks(J, i, j)
                assert J_sub[2 * i + j].signature() == J_ij_ext.signature()


def test_postive_restricted_extract_none():
    cell = ufl.triangle
    d = cell.topological_dimension()
    domain = ufl.Mesh(FiniteElement("Lagrange", cell, 1, (d,), ufl.identity_pullback, ufl.H1))
    el_u = FiniteElement("Lagrange", cell, 2, (d,), ufl.identity_pullback, ufl.H1)
    el_p = FiniteElement("Lagrange", cell, 1, (), ufl.identity_pullback, ufl.H1)
    V = ufl.FunctionSpace(domain, el_u)
    Q = ufl.FunctionSpace(domain, el_p)
    W = ufl.MixedFunctionSpace(V, Q)
    u, p = ufl.TrialFunctions(W)
    v, q = ufl.TestFunctions(W)
    a = (
        ufl.inner(ufl.grad(u), ufl.grad(v)) * ufl.dx
        + ufl.div(u) * q * ufl.dx
        + ufl.div(v) * p * ufl.dx
    )
    a += ufl.inner(u("+"), v("+")) * ufl.dS
    a_blocks = ufl.extract_blocks(a)
    assert a_blocks[1][1] is None



# Copyright (C) 2005-2007 Anders Logg
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# The bilinear form a(v, u) and linear form L(v) for
# tensor-weighted Poisson's equation.
from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dx,
    grad,
    inner,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1, L2

P1 = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
P0 = FiniteElement("Discontinuous Lagrange", triangle, 0, (2, 2), identity_pullback, L2)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
p1_space = FunctionSpace(domain, P1)
p0_space = FunctionSpace(domain, P0)

v = TestFunction(p1_space)
u = TrialFunction(p1_space)
C = Coefficient(p0_space)

a = inner(grad(v), C * grad(u)) * dx



"""Test scratch.

This is a template file you can copy when making a new test case.
Begin by copying this file to a filename matching test_*.py.
The tests in the file will then automatically be run by ./test.py.
Next look at the TODO markers below for places to edit.
"""

import warnings

from ufl import (
    Coefficient,
    FunctionSpace,
    Identity,
    Mesh,
    TestFunction,
    as_matrix,
    as_tensor,
    as_vector,
    dx,
    grad,
    indices,
    inner,
    outer,
    triangle,
)
from ufl.classes import FixedIndex, FormArgument, Grad, Indexed, ListTensor, Zero
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1
from ufl.tensors import as_scalar, unit_indexed_tensor, unwrap_list_tensor


class MockForwardAD:
    def __init__(self):
        self._w = ()
        self._v = ()

        class Obj:
            def __init__(self):
                self._data = {}

        self._cd = Obj()

    def grad(self, g):
        # If we hit this type, it has already been propagated
        # to a coefficient (or grad of a coefficient), # FIXME: Assert this!
        # so we need to take the gradient of the variation or return zero.
        # Complications occur when dealing with derivatives w.r.t. single components...

        # Figure out how many gradients are around the inner terminal
        ngrads = 0
        o = g
        while isinstance(o, Grad):
            (o,) = o.ufl_operands
            ngrads += 1
        if not isinstance(o, FormArgument):
            raise ValueError("Expecting gradient of a FormArgument, not %s" % repr(o))

        def apply_grads(f):
            if not isinstance(f, FormArgument):
                print(("," * 60))
                print(f)
                print(o)
                print(g)
                print(("," * 60))
                raise ValueError("What?")
            for i in range(ngrads):
                f = Grad(f)
            return f

        # Find o among all w without any indexing, which makes this easy
        for w, v in zip(self._w, self._v):
            if o == w and isinstance(v, FormArgument):
                # Case: d/dt [w + t v]
                return (g, apply_grads(v))

        # If o is not among coefficient derivatives, return do/dw=0
        gprimesum = Zero(g.ufl_shape)

        def analyse_variation_argument(v):
            # Analyse variation argument
            if isinstance(v, FormArgument):
                # Case: d/dt [w[...] + t v]
                vval, vcomp = v, ()
            elif isinstance(v, Indexed):
                # Case: d/dt [w + t v[...]]
                # Case: d/dt [w[...] + t v[...]]
                vval, vcomp = v.ufl_operands
                vcomp = tuple(vcomp)
            else:
                raise ValueError("Expecting argument or component of argument.")
            if not all(isinstance(k, FixedIndex) for k in vcomp):
                raise ValueError("Expecting only fixed indices in variation.")
            return vval, vcomp

        def compute_gprimeterm(ngrads, vval, vcomp, wshape, wcomp):
            # Apply gradients directly to argument vval,
            # and get the right indexed scalar component(s)
            kk = indices(ngrads)
            Dvkk = apply_grads(vval)[vcomp + kk]
            # Place scalar component(s) Dvkk into the right tensor positions
            if wshape:
                Ejj, jj = unit_indexed_tensor(wshape, wcomp)
            else:
                Ejj, jj = 1, ()
            gprimeterm = as_tensor(Ejj * Dvkk, jj + kk)
            return gprimeterm

        # Accumulate contributions from variations in different components
        for w, v in zip(self._w, self._v):
            # Analyse differentiation variable coefficient
            if isinstance(w, FormArgument):
                if not w == o:
                    continue
                wshape = w.ufl_shape

                if isinstance(v, FormArgument):
                    # Case: d/dt [w + t v]
                    return (g, apply_grads(v))

                elif isinstance(v, ListTensor):
                    # Case: d/dt [w + t <...,v,...>]
                    for wcomp, vsub in unwrap_list_tensor(v):
                        if not isinstance(vsub, Zero):
                            vval, vcomp = analyse_variation_argument(vsub)
                            gprimesum = gprimesum + compute_gprimeterm(
                                ngrads, vval, vcomp, wshape, wcomp
                            )

                else:
                    if wshape != ():
                        raise ValueError("Expecting scalar coefficient in this branch.")
                    # Case: d/dt [w + t v[...]]
                    wval, wcomp = w, ()

                    vval, vcomp = analyse_variation_argument(v)
                    gprimesum = gprimesum + compute_gprimeterm(ngrads, vval, vcomp, wshape, wcomp)

            elif isinstance(
                w, Indexed
            ):  # This path is tested in unit tests, but not actually used?
                # Case: d/dt [w[...] + t v[...]]
                # Case: d/dt [w[...] + t v]
                wval, wcomp = w.ufl_operands
                if not wval == o:
                    continue
                assert isinstance(wval, FormArgument)
                if not all(isinstance(k, FixedIndex) for k in wcomp):
                    raise ValueError("Expecting only fixed indices in differentiation variable.")
                wshape = wval.ufl_shape

                vval, vcomp = analyse_variation_argument(v)
                gprimesum = gprimesum + compute_gprimeterm(ngrads, vval, vcomp, wshape, wcomp)

            else:
                raise ValueError("Expecting coefficient or component of coefficient.")

        # FIXME: Handle other coefficient derivatives: oprimes = self._cd._data.get(o)

        if 0:
            oprimes = self._cd._data.get(o)
            if oprimes is None:
                if self._cd._data:
                    # TODO: Make it possible to silence this message in particular?
                    #       It may be good to have for debugging...
                    warnings.warn("Assuming d{%s}/d{%s} = 0." % (o, self._w))
            else:
                # Make sure we have a tuple to match the self._v tuple
                if not isinstance(oprimes, tuple):
                    oprimes = (oprimes,)
                    if len(oprimes) != len(self._v):
                        raise ValueError(
                            "Got a tuple of arguments, "
                            "expecting a matching tuple of coefficient derivatives."
                        )

                # Compute dg/dw_j = dg/dw_h : v.
                # Since we may actually have a tuple of oprimes and vs in a
                # 'mixed' space, sum over them all to get the complete inner
                # product. Using indices to define a non-compound inner product.
                for oprime, v in zip(oprimes, self._v):
                    raise ValueError("FIXME: Figure out how to do this with ngrads")
                    so, oi = as_scalar(oprime)
                    rv = len(v.ufl_shape)
                    oi1 = oi[:-rv]
                    oi2 = oi[-rv:]
                    prod = so * v[oi2]
                    if oi1:
                        gprimesum += as_tensor(prod, oi1)
                    else:
                        gprimesum += prod

        return (g, gprimesum)


def test_unit_tensor(self):
    E2_1, ii = unit_indexed_tensor((2,), (1,))
    E3_1, ii = unit_indexed_tensor((3,), (1,))
    E22_10, ii = unit_indexed_tensor((2, 2), (1, 0))
    # TODO: Evaluate and assert values


def test_unwrap_list_tensor(self):
    lt = as_tensor((1, 2))
    expected = [
        ((0,), 1),
        ((1,), 2),
    ]
    comp = unwrap_list_tensor(lt)
    assert comp == expected

    lt = as_tensor(((1, 2), (3, 4)))
    expected = [
        ((0, 0), 1),
        ((0, 1), 2),
        ((1, 0), 3),
        ((1, 1), 4),
    ]
    comp = unwrap_list_tensor(lt)
    assert comp == expected

    lt = as_tensor((((1, 2), (3, 4)), ((11, 12), (13, 14))))
    expected = [
        ((0, 0, 0), 1),
        ((0, 0, 1), 2),
        ((0, 1, 0), 3),
        ((0, 1, 1), 4),
        ((1, 0, 0), 11),
        ((1, 0, 1), 12),
        ((1, 1, 0), 13),
        ((1, 1, 1), 14),
    ]
    comp = unwrap_list_tensor(lt)
    assert comp == expected


def test__forward_coefficient_ad__grad_of_scalar_coefficient(self):
    U = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, U)
    u = Coefficient(space)
    du = TestFunction(space)

    mad = MockForwardAD()
    mad._w = (u,)
    mad._v = (du,)

    # Simple grad(coefficient) -> grad(variation)
    f = grad(u)
    df = grad(du)
    g, dg = mad.grad(f)
    assert g == f
    assert dg == df

    # Simple grad(grad(coefficient)) -> grad(grad(variation))
    f = grad(grad(u))
    df = grad(grad(du))
    g, dg = mad.grad(f)
    assert g == f
    assert dg == df


def test__forward_coefficient_ad__grad_of_vector_coefficient(self):
    V = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    v = Coefficient(space)
    dv = TestFunction(space)

    mad = MockForwardAD()
    mad._w = (v,)
    mad._v = (dv,)

    # Simple grad(coefficient) -> grad(variation)
    f = grad(v)
    df = grad(dv)
    g, dg = mad.grad(f)
    assert g == f
    assert dg == df

    # Simple grad(grad(coefficient)) -> grad(grad(variation))
    f = grad(grad(v))
    df = grad(grad(dv))
    g, dg = mad.grad(f)
    assert g == f
    assert dg == df


def test__forward_coefficient_ad__grad_of_vector_coefficient__with_component_variation(self):
    V = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    v = Coefficient(space)
    dv = TestFunction(space)

    mad = MockForwardAD()

    # Component of variation:
    # grad(grad(c))[0,...] -> grad(grad(dc))[1,...]
    mad._w = (v[0],)
    mad._v = (dv[1],)
    f = grad(v)
    df = grad(as_vector((dv[1], 0)))  # Mathematically this would be the natural result
    j, k = indices(2)
    df = as_tensor(
        Identity(2)[0, j] * grad(dv)[1, k], (j, k)
    )  # Actual representation should have grad right next to dv
    g, dg = mad.grad(f)
    if 0:
        print(("\nf    ", f))
        print(("df   ", df))
        print(("g    ", g))
        print(("dg   ", dg))
    assert f.ufl_shape == df.ufl_shape
    assert g.ufl_shape == f.ufl_shape
    assert dg.ufl_shape == df.ufl_shape
    assert g == f
    self.assertEqual((inner(dg, dg) * dx).signature(), (inner(df, df) * dx).signature())
    # assert dg == df # Expected to fail because of different index numbering

    # Multiple components of variation:
    # grad(grad(c))[0,1,:,:] -> grad(grad(dc))[1,0,:,:]
    mad._w = (v[0], v[1])
    mad._v = (dv[1], dv[0])
    f = grad(v)
    # Mathematically this would be the natural result:
    df = grad(as_vector((dv[1], dv[0])))
    # Actual representation should have grad right next to dv:
    j0, k0 = indices(2)
    j1, k1 = indices(2)  # Using j0,k0 for both terms gives different signature
    df = as_tensor(Identity(2)[0, j0] * grad(dv)[1, k0], (j0, k0)) + as_tensor(
        Identity(2)[1, j1] * grad(dv)[0, k1], (j1, k1)
    )
    g, dg = mad.grad(f)
    print(("\nf    ", f))
    print(("df   ", df))
    print(("g    ", g))
    print(("dg   ", dg))
    assert f.ufl_shape == df.ufl_shape
    assert g.ufl_shape == f.ufl_shape
    assert dg.ufl_shape == df.ufl_shape
    assert g == f
    self.assertEqual((inner(dg, dg) * dx).signature(), (inner(df, df) * dx).signature())
    # assert dg == df # Expected to fail because of different index numbering


def test__forward_coefficient_ad__grad_of_vector_coefficient__with_component_variation_in_list(
    self,
):
    V = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    v = Coefficient(space)
    dv = TestFunction(space)

    mad = MockForwardAD()

    # Component of variation:
    # grad(grad(c))[0,...] -> grad(grad(dc))[1,...]
    mad._w = (v,)
    mad._v = (as_vector((dv[1], 0)),)
    f = grad(v)
    df = grad(as_vector((dv[1], 0)))  # Mathematically this would be the natural result
    j, k = indices(2)
    df = as_tensor(
        Identity(2)[0, j] * grad(dv)[1, k], (j, k)
    )  # Actual representation should have grad right next to dv
    g, dg = mad.grad(f)
    if 0:
        print(("\nf    ", f))
        print(("df   ", df))
        print(("g    ", g))
        print(("dg   ", dg))
    assert f.ufl_shape == df.ufl_shape
    assert g.ufl_shape == f.ufl_shape
    assert dg.ufl_shape == df.ufl_shape
    assert g == f
    self.assertEqual((inner(dg, dg) * dx).signature(), (inner(df, df) * dx).signature())
    # assert dg == df # Expected to fail because of different index numbering

    # Multiple components of variation:
    # grad(grad(c))[0,1,:,:] -> grad(grad(dc))[1,0,:,:]
    mad._w = (v,)
    mad._v = (as_vector((dv[1], dv[0])),)
    f = grad(v)
    # Mathematically this would be the natural result:
    df = grad(as_vector((dv[1], dv[0])))
    # Actual representation should have grad right next to dv:
    j0, k0 = indices(2)
    j1, k1 = indices(2)  # Using j0,k0 for both terms gives different signature
    df = as_tensor(Identity(2)[0, j0] * grad(dv)[1, k0], (j0, k0)) + as_tensor(
        Identity(2)[1, j1] * grad(dv)[0, k1], (j1, k1)
    )
    g, dg = mad.grad(f)
    print(("\nf    ", f))
    print(("df   ", df))
    print(("g    ", g))
    print(("dg   ", dg))
    assert f.ufl_shape == df.ufl_shape
    assert g.ufl_shape == f.ufl_shape
    assert dg.ufl_shape == df.ufl_shape
    assert g == f
    self.assertEqual((inner(dg, dg) * dx).signature(), (inner(df, df) * dx).signature())
    # assert dg == df # Expected to fail because of different index numbering


def test__forward_coefficient_ad__grad_of_tensor_coefficient(self):
    W = FiniteElement("Lagrange", triangle, 1, (2, 2), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, W)
    w = Coefficient(space)
    dw = TestFunction(space)

    mad = MockForwardAD()
    mad._w = (w,)
    mad._v = (dw,)

    # Simple grad(coefficient) -> grad(variation)
    f = grad(w)
    df = grad(dw)
    g, dg = mad.grad(f)
    assert g == f
    assert dg == df

    # Simple grad(grad(coefficient)) -> grad(grad(variation))
    f = grad(grad(w))
    df = grad(grad(dw))
    g, dg = mad.grad(f)
    assert g == f
    assert dg == df


def test__forward_coefficient_ad__grad_of_tensor_coefficient__with_component_variation(self):
    W = FiniteElement("Lagrange", triangle, 1, (2, 2), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, W)
    w = Coefficient(space)
    dw = TestFunction(space)

    mad = MockForwardAD()

    # Component of variation:
    # grad(grad(c))[0,...] -> grad(grad(dc))[1,...]
    wc = (1, 0)
    dwc = (0, 1)
    mad._w = (w[wc],)
    mad._v = (dw[dwc],)
    f = grad(w)
    df = grad(as_matrix(((0, 0), (dw[dwc], 0))))  # Mathematically this is it.
    i, j, k = indices(3)
    E = outer(Identity(2)[wc[0], i], Identity(2)[wc[1], j])
    Ddw = grad(dw)[dwc + (k,)]
    df = as_tensor(E * Ddw, (i, j, k))  # Actual representation should have grad next to dv
    g, dg = mad.grad(f)
    if 0:
        print(("\nf    ", f))
        print(("df   ", df))
        print(("g    ", g))
        print(("dg   ", dg))
    assert f.ufl_shape == df.ufl_shape
    assert g.ufl_shape == f.ufl_shape
    assert dg.ufl_shape == df.ufl_shape
    assert g == f
    self.assertEqual((inner(dg, dg) * dx).signature(), (inner(df, df) * dx).signature())
    # assert dg == df # Expected to fail because of different index numbering



"""The Form class."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009-2011.
# Modified by Massimiliano Leoni, 2016.
# Modified by Cecile Daversin-Catty, 2018.
# Modified by Nacime Bouziani, 2020.
# Modified by Jørgen S. Dokken 2023.

import numbers
import warnings
from collections import defaultdict
from itertools import chain

from ufl.checks import is_scalar_constant_expression
from ufl.constant import Constant
from ufl.constantvalue import Zero
from ufl.core.expr import Expr, ufl_err_str
from ufl.core.ufl_type import UFLType, ufl_type
from ufl.domain import extract_unique_domain, sort_domains
from ufl.equation import Equation
from ufl.integral import Integral
from ufl.utils.counted import Counted
from ufl.utils.sorting import sorted_by_count

# Export list for ufl.classes
__all_classes__ = ["Form", "BaseForm", "ZeroBaseForm"]

# --- The Form class, representing a complete variational form or functional ---


def _sorted_integrals(integrals):
    """Sort integrals for a stable signature computation.

    Sort integrals by domain id, integral type, subdomain id for a more
    stable signature computation.
    """
    # Group integrals in multilevel dict by keys
    # [domain][integral_type][subdomain_id]
    integrals_dict = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
    for integral in integrals:
        d = integral.ufl_domain()
        if d is None:
            raise ValueError(
                "Each integral in a form must have a uniquely defined integration domain."
            )
        it = integral.integral_type()
        si = integral.subdomain_id()
        integrals_dict[d][it][si] += [integral]

    all_integrals = []

    def keyfunc(item):
        if isinstance(item, numbers.Integral):
            sid_int = item
        else:
            # As subdomain ids can be either int or tuples, we need to compare them
            sid_int = tuple(-1 if i == "otherwise" else i for i in item)
        return (type(item).__name__, sid_int)

    # Order integrals canonically to increase signature stability
    for d in sort_domains(integrals_dict):
        for it in sorted(integrals_dict[d]):  # str is sortable
            for si in sorted(integrals_dict[d][it], key=keyfunc):
                unsorted_integrals = integrals_dict[d][it][si]
                # TODO: At this point we could order integrals by
                #       metadata and integrand, or even add the
                #       integrands with the same metadata. This is done
                #       in accumulate_integrands_with_same_metadata in
                #       algorithms/domain_analysis.py and would further
                #       increase the signature stability.
                all_integrals.extend(unsorted_integrals)
                # integrals_dict[d][it][si] = unsorted_integrals

    return tuple(all_integrals)  # integrals_dict


@ufl_type()
class BaseForm(object, metaclass=UFLType):
    """Description of an object containing arguments."""

    # Slots is kept empty to enable multiple inheritance with other
    # classes
    __slots__ = ()
    _ufl_is_abstract_ = True
    _ufl_required_methods_ = ("_analyze_form_arguments", "_analyze_domains", "ufl_domains")

    def __init__(self):
        """Initialise."""
        # Internal variables for caching form argument/coefficient data
        self._arguments = None
        self._coefficients = None

    # --- Accessor interface ---
    def arguments(self):
        """Return all ``Argument`` objects found in form."""
        if self._arguments is None:
            self._analyze_form_arguments()
        return self._arguments

    def coefficients(self):
        """Return all ``Coefficient`` objects found in form."""
        if self._coefficients is None:
            self._analyze_form_arguments()
        return self._coefficients

    def ufl_domain(self):
        """Return the single geometric integration domain occuring in the base form.

        Fails if multiple domains are found.
        """
        try:
            (domain,) = set(self.ufl_domains())
        except ValueError:
            raise ValueError("%s must have exactly one domain." % type(self).__name__)
        # Return the one and only domain
        return domain

    # --- Operator implementations ---

    def __eq__(self, other):
        """Delayed evaluation of the == operator.

        Just 'lhs_form == rhs_form' gives an Equation,
        while 'bool(lhs_form == rhs_form)' delegates
        to lhs_form.equals(rhs_form).
        """
        return Equation(self, other)

    def __radd__(self, other):
        """Add."""
        # Ordering of form additions make no difference
        return self.__add__(other)

    def __add__(self, other):
        """Add."""
        if isinstance(other, numbers.Number) and other == 0:
            # Allow adding 0 or 0.0 as a no-op, needed for sum([a,b])
            return self
        elif isinstance(other, Zero):
            # Allow adding ufl Zero as a no-op, needed for sum([a,b])
            return self

        elif isinstance(other, ZeroBaseForm):
            # Simplify addition with ZeroBaseForm
            return self

        # For `ZeroBaseForm(...) + B` with B a BaseForm.
        # We could overwrite ZeroBaseForm.__add__ but that implies
        # duplicating cases with `0` and `ufl.Zero`.
        elif isinstance(self, ZeroBaseForm):
            # Simplify addition with ZeroBaseForm
            return other

        elif isinstance(other, BaseForm):
            # Add integrals from both forms
            return FormSum((self, 1), (other, 1))

        else:
            # Let python protocols do their job if we don't handle it
            return NotImplemented

    def __sub__(self, other):
        """Subtract other form from this one."""
        return self + (-other)

    def __rsub__(self, other):
        """Subtract this form from other."""
        return other + (-self)

    def __neg__(self):
        """Negate all integrals in form.

        This enables the handy "-form" syntax for e.g. the
        linearized system (J, -F) from a nonlinear form F.
        """
        if isinstance(self, ZeroBaseForm):
            # `-` doesn't change anything for ZeroBaseForm.
            # This also facilitates simplifying FormSum containing ZeroBaseForm objects.
            return self
        return FormSum((self, -1))

    def __rmul__(self, scalar):
        """Multiply all integrals in form with constant scalar value."""
        # This enables the handy "0*form" or "dt*form" syntax
        if is_scalar_constant_expression(scalar):
            return FormSum((self, scalar))
        return NotImplemented

    def __mul__(self, coefficient):
        """Take the action of this form on the given coefficient."""
        if isinstance(coefficient, Expr):
            from ufl.formoperators import action

            return action(self, coefficient)
        return NotImplemented

    def __ne__(self, other):
        """Immediately evaluate the != operator (as opposed to the == operator)."""
        return not self.equals(other)

    def __call__(self, x):
        """Take the action of this form on ``x``."""
        from ufl.formoperators import action

        return action(self, x)

    def _ufl_compute_hash_(self):
        """Compute the hash."""
        # Ensure compatibility with MultiFunction
        # `hash(self)` will call the `__hash__` method of the subclass.
        return hash(self)

    def _ufl_expr_reconstruct_(self, *operands):
        """Return a new object of the same type with new operands."""
        return type(self)(*operands)

    __matmul__ = __mul__


@ufl_type()
class Form(BaseForm):
    """Description of a weak form consisting of a sum of integrals over subdomains."""

    __slots__ = (
        # --- List of Integral objects (a Form is a sum of these
        # Integrals, everything else is derived)
        "_integrals",
        # --- Internal variables for caching various data
        "_integration_domains",
        "_domain_numbering",
        "_subdomain_data",
        "_arguments",
        "_base_form_operators",
        "_coefficients",
        "_coefficient_numbering",
        "_constants",
        "_constant_numbering",
        "_terminal_numbering",
        "_hash",
        "_signature",
        # --- Dict that external frameworks can place framework-specific
        #     data in to be carried with the form
        #     Never use this internally in ufl!
        "_cache",
    )

    def __init__(self, integrals):
        """Initialise."""
        BaseForm.__init__(self)
        # Basic input checking (further compatibilty analysis happens
        # later)
        if not all(isinstance(itg, Integral) for itg in integrals):
            raise ValueError("Expecting list of integrals.")

        # Store integrals sorted canonically to increase signature
        # stability
        self._integrals = _sorted_integrals(integrals)

        # Internal variables for caching domain data
        self._integration_domains = None
        self._domain_numbering = None

        # Internal variables for caching subdomain data
        self._subdomain_data = None

        # Internal variables for caching form argument data
        self._coefficients = None
        self._coefficient_numbering = None
        self._constant_numbering = None
        self._terminal_numbering = None

        # Internal variables for caching base form operator data
        self._base_form_operators = None

        from ufl.algorithms.analysis import extract_constants

        self._constants = extract_constants(self)

        # Internal variables for caching of hash and signature after
        # first request
        self._hash = None
        self._signature = None

        # Never use this internally in ufl!
        self._cache = {}

    # --- Accessor interface ---

    def integrals(self):
        """Return a sequence of all integrals in form."""
        return self._integrals

    def integrals_by_type(self, integral_type):
        """Return a sequence of all integrals with a particular domain type."""
        return tuple(
            integral for integral in self.integrals() if integral.integral_type() == integral_type
        )

    def integrals_by_domain(self, domain):
        """Return a sequence of all integrals with a particular integration domain."""
        return tuple(integral for integral in self.integrals() if integral.ufl_domain() == domain)

    def empty(self):
        """Returns whether the form has no integrals."""
        return self.integrals() == ()

    def ufl_domains(self):
        """Return the geometric integration domains occuring in the form.

        NB! This does not include domains of coefficients defined on
        other meshes.

        The return type is a tuple even if only a single domain exists.
        """
        if self._integration_domains is None:
            self._analyze_domains()
        return self._integration_domains

    def ufl_cell(self):
        """Return the single cell this form is defined on.

        Fails if multiple cells are found.
        """
        return self.ufl_domain().ufl_cell()

    def geometric_dimension(self):
        """Return the geometric dimension shared by all domains and functions in this form."""
        gdims = tuple(set(domain.geometric_dimension() for domain in self.ufl_domains()))
        if len(gdims) != 1:
            raise ValueError(
                "Expecting all domains and functions in a form "
                f"to share geometric dimension, got {tuple(sorted(gdims))}"
            )
        return gdims[0]

    def domain_numbering(self):
        """Return a contiguous numbering of domains in a mapping ``{domain:number}``."""
        if self._domain_numbering is None:
            self._analyze_domains()
        return self._domain_numbering

    def subdomain_data(self):
        """Returns a mapping on the form ``{domain:{integral_type: subdomain_data}}``."""
        if self._subdomain_data is None:
            self._analyze_subdomain_data()
        return self._subdomain_data

    def coefficients(self):
        """Return all ``Coefficient`` objects found in form."""
        if self._coefficients is None:
            self._analyze_form_arguments()
        return self._coefficients

    def base_form_operators(self):
        """Return all ``BaseFormOperator`` objects found in form."""
        if self._base_form_operators is None:
            self._analyze_base_form_operators()
        return self._base_form_operators

    def coefficient_numbering(self):
        """Return a contiguous numbering of coefficients in a mapping ``{coefficient:number}``."""
        # cyclic import
        from ufl.coefficient import Coefficient

        if self._coefficient_numbering is None:
            self._coefficient_numbering = {
                expr: num
                for expr, num in self.terminal_numbering().items()
                if isinstance(expr, Coefficient)
            }
        return self._coefficient_numbering

    def constants(self):
        """Get constants."""
        return self._constants

    def constant_numbering(self):
        """Return a contiguous numbering of constants in a mapping ``{constant:number}``."""
        if self._constant_numbering is None:
            self._constant_numbering = {
                expr: num
                for expr, num in self.terminal_numbering().items()
                if isinstance(expr, Constant)
            }
        return self._constant_numbering

    def terminal_numbering(self):
        """Return a contiguous numbering for all counted objects in the form.

        The returned object is mapping from terminal to its number (an integer).

        The numbering is computed per type so :class:`Coefficient`s,
        :class:`Constant`s, etc will each be numbered from zero.
        """
        # cyclic import
        from ufl.algorithms.analysis import extract_type

        if self._terminal_numbering is None:
            exprs_by_type = defaultdict(set)
            for counted_expr in extract_type(self, Counted):
                exprs_by_type[counted_expr._counted_class].add(counted_expr)

            numbering = {}
            for exprs in exprs_by_type.values():
                for i, expr in enumerate(sorted_by_count(exprs)):
                    numbering[expr] = i
            self._terminal_numbering = numbering
        return self._terminal_numbering

    def signature(self):
        """Signature for use with jit cache (independent of incidental numbering of indices etc)."""
        if self._signature is None:
            self._compute_signature()
        return self._signature

    # --- Operator implementations ---

    def __hash__(self):
        """Hash."""
        if self._hash is None:
            self._hash = hash(tuple(hash(itg) for itg in self.integrals()))
        return self._hash

    def __ne__(self, other):
        """Immediate evaluation of the != operator (as opposed to the == operator)."""
        return not self.equals(other)

    def equals(self, other):
        """Evaluate ``bool(lhs_form == rhs_form)``."""
        if type(other) is not Form:
            return False
        if len(self._integrals) != len(other._integrals):
            return False
        if hash(self) != hash(other):
            return False
        return all(a == b for a, b in zip(self._integrals, other._integrals))

    def __radd__(self, other):
        """Add."""
        # Ordering of form additions make no difference
        return self.__add__(other)

    def __add__(self, other):
        """Add."""
        if isinstance(other, Form):
            # Add integrals from both forms
            return Form(list(chain(self.integrals(), other.integrals())))

        if isinstance(other, ZeroBaseForm):
            # Simplify addition with ZeroBaseForm
            return self

        elif isinstance(other, BaseForm):
            # Create form sum if form is of other type
            return FormSum((self, 1), (other, 1))

        elif isinstance(other, (int, float)) and other == 0:
            # Allow adding 0 or 0.0 as a no-op, needed for sum([a,b])
            return self

        elif isinstance(other, Zero) and not (other.ufl_shape or other.ufl_free_indices):
            # Allow adding ufl Zero as a no-op, needed for sum([a,b])
            return self

        else:
            # Let python protocols do their job if we don't handle it
            return NotImplemented

    def __sub__(self, other):
        """Subtract other form from this one."""
        return self + (-other)

    def __rsub__(self, other):
        """Subtract this form from other."""
        return other + (-self)

    def __neg__(self):
        """Negate all integrals in form.

        This enables the handy "-form" syntax for e.g. the
        linearized system (J, -F) from a nonlinear form F.
        """
        return Form([-itg for itg in self.integrals()])

    def __rmul__(self, scalar):
        """Multiply all integrals in form with constant scalar value."""
        # This enables the handy "0*form" or "dt*form" syntax
        if is_scalar_constant_expression(scalar):
            return Form([scalar * itg for itg in self.integrals()])
        return NotImplemented

    def __mul__(self, coefficient):
        """UFL form operator: Take the action of this form on the given coefficient."""
        if isinstance(coefficient, Expr):
            from ufl.formoperators import action

            return action(self, coefficient)
        return NotImplemented

    def __call__(self, *args, **kwargs):
        """UFL form operator: Evaluate form by replacing arguments and coefficients.

        Replaces form.arguments() with given positional arguments in
        same number and ordering. Number of positional arguments must
        be 0 or equal to the number of Arguments in the form.

        The optional keyword argument coefficients can be set to a dict
        to replace Coefficients with expressions of matching shapes.

        Example:
          V = FiniteElement("CG", triangle, 1)
          v = TestFunction(V)
          u = TrialFunction(V)
          f = Coefficient(V)
          g = Coefficient(V)
          a = g*inner(grad(u), grad(v))*dx
          M = a(f, f, coefficients={ g: 1 })

        Is equivalent to M == grad(f)**2*dx.
        """
        repdict = {}

        if args:
            arguments = self.arguments()
            if len(arguments) != len(args):
                raise ValueError(f"Need {len(arguments)} arguments to form(), got {len(args)}.")
            repdict.update(zip(arguments, args))

        coefficients = kwargs.pop("coefficients", None)
        if kwargs:
            raise ValueError(f"Unknown kwargs {list(kwargs)}")

        if coefficients is not None:
            coeffs = self.coefficients()
            for f in coefficients:
                if f in coeffs:
                    repdict[f] = coefficients[f]
                else:
                    warnings.warn("Coefficient %s is not in form." % ufl_err_str(f))
        if repdict:
            from ufl.formoperators import replace

            return replace(self, repdict)
        else:
            return self

    __matmul__ = __mul__

    # --- String conversion functions, for UI purposes only ---

    def __str__(self):
        """Compute shorter string representation of form. This can be huge for complicated forms."""
        # Warning used for making sure we don't use this in the general pipeline:
        # warning("Calling str on form is potentially expensive and
        # should be avoided except during debugging.") Not caching this
        # because it can be huge
        s = "\n  +  ".join(str(itg) for itg in self.integrals())
        return s or "<empty Form>"

    def __repr__(self):
        """Compute repr string of form. This can be huge for complicated forms."""
        # Warning used for making sure we don't use this in the general pipeline:
        # warning("Calling repr on form is potentially expensive and
        # should be avoided except during debugging.") Not caching this
        # because it can be huge
        itgs = ", ".join(repr(itg) for itg in self.integrals())
        r = "Form([" + itgs + "])"
        return r

    # --- Analysis functions, precomputation and caching of various quantities

    def _analyze_domains(self):
        """Analyze domains."""
        from ufl.domain import join_domains, sort_domains

        # Collect unique integration domains
        integration_domains = join_domains([itg.ufl_domain() for itg in self._integrals])

        # Make canonically ordered list of the domains
        self._integration_domains = sort_domains(integration_domains)

        # TODO: Not including domains from coefficients and arguments
        # here, may need that later
        self._domain_numbering = dict((d, i) for i, d in enumerate(self._integration_domains))

    def _analyze_subdomain_data(self):
        """Analyze subdomain data."""
        integration_domains = self.ufl_domains()
        integrals = self.integrals()

        # Make clear data structures to collect subdomain data in
        subdomain_data = {}
        for domain in integration_domains:
            subdomain_data[domain] = {}

        for integral in integrals:
            # Get integral properties
            domain = integral.ufl_domain()
            it = integral.integral_type()
            sd = integral.subdomain_data()

            # Collect subdomain data
            if subdomain_data[domain].get(it) is None:
                subdomain_data[domain][it] = [sd]
            else:
                subdomain_data[domain][it].append(sd)
        self._subdomain_data = subdomain_data

    def _analyze_form_arguments(self):
        """Analyze which Argument and Coefficient objects can be found in the form."""
        from ufl.algorithms.analysis import extract_arguments_and_coefficients

        arguments, coefficients = extract_arguments_and_coefficients(self)

        # Define canonical numbering of arguments and coefficients
        self._arguments = tuple(sorted(set(arguments), key=lambda x: x.number()))
        self._coefficients = tuple(sorted(set(coefficients), key=lambda x: x.count()))

    def _analyze_base_form_operators(self):
        """Analyze which BaseFormOperator objects can be found in the form."""
        from ufl.algorithms.analysis import extract_base_form_operators

        base_form_ops = extract_base_form_operators(self)
        self._base_form_operators = tuple(sorted(base_form_ops, key=lambda x: x.count()))

    def _compute_renumbering(self):
        """Compute renumbering."""
        # Include integration domains and coefficients in renumbering
        dn = self.domain_numbering()
        tn = self.terminal_numbering()
        renumbering = {}
        renumbering.update(dn)
        renumbering.update(tn)

        # Add domains of coefficients, these may include domains not
        # among integration domains
        k = len(dn)
        for c in self.coefficients():
            d = extract_unique_domain(c)
            if d is not None and d not in renumbering:
                renumbering[d] = k
                k += 1

        # Add domains of arguments, these may include domains not
        # among integration domains
        for a in self._arguments:
            d = a.ufl_function_space().ufl_domain()
            if d is not None and d not in renumbering:
                renumbering[d] = k
                k += 1

        # Add domains of constants, these may include domains not
        # among integration domains
        for c in self._constants:
            d = extract_unique_domain(c)
            if d is not None and d not in renumbering:
                renumbering[d] = k
                k += 1

        return renumbering

    def _compute_signature(self):
        """Compute signature."""
        from ufl.algorithms.signature import compute_form_signature

        self._signature = compute_form_signature(self, self._compute_renumbering())


def as_form(form):
    """Convert to form if not a form, otherwise return form."""
    if not isinstance(form, BaseForm) and form != 0:
        raise ValueError(f"Unable to convert object to a UFL form: {ufl_err_str(form)}")
    return form


@ufl_type()
class FormSum(BaseForm):
    """Form sum.

    Description of a weighted sum of variational forms and form-like objects
    components is the list of Forms to be summed
    arg_weights is a list of tuples of component index and weight
    """

    __slots__ = (
        "_arguments",
        "_coefficients",
        "_weights",
        "_components",
        "ufl_operands",
        "_domains",
        "_domain_numbering",
        "_hash",
    )
    _ufl_required_methods_ = "_analyze_form_arguments"

    def __new__(cls, *args, **kwargs):
        """Create a new FormSum."""
        # All the components are `ZeroBaseForm`
        if all(component == 0 for component, _ in args):
            # Assume that the arguments of all the components have
            # consistent with each other  and select the first one to
            # define the arguments of `ZeroBaseForm`.
            # This might not always be true but `ZeroBaseForm`'s
            # arguments are not checked anywhere because we can't
            # reliably always infer them.
            ((arg, _), *_) = args
            arguments = arg.arguments()
            return ZeroBaseForm(arguments)

        return super(FormSum, cls).__new__(cls)

    def __init__(self, *components):
        """Initialise."""
        BaseForm.__init__(self)

        # Remove `ZeroBaseForm` components
        filtered_components = [(component, w) for component, w in components if component != 0]

        weights = []
        full_components = []
        for component, w in filtered_components:
            if isinstance(component, FormSum):
                full_components.extend(component.components())
                weights.extend([w * wc for wc in component.weights()])
            else:
                full_components.append(component)
                weights.append(w)

        self._arguments = None
        self._coefficients = None
        self._domains = None
        self._domain_numbering = None
        self._hash = None
        self._weights = weights
        self._components = full_components
        self._sum_variational_components()
        self.ufl_operands = self._components

    def components(self):
        """Get components."""
        return self._components

    def weights(self):
        """Get weights."""
        return self._weights

    def _sum_variational_components(self):
        """Sum variational components."""
        var_forms = None
        other_components = []
        new_weights = []
        for i, component in enumerate(self._components):
            if isinstance(component, Form):
                if var_forms:
                    var_forms = var_forms + (self._weights[i] * component)
                else:
                    var_forms = self._weights[i] * component
            else:
                other_components.append(component)
                new_weights.append(self._weights[i])
        if var_forms:
            other_components.insert(0, var_forms)
            new_weights.insert(0, 1)
        self._components = other_components
        self._weights = new_weights

    def _analyze_form_arguments(self):
        """Return all ``Argument`` objects found in form."""
        arguments = []
        coefficients = []
        for component in self._components:
            arguments.extend(component.arguments())
            coefficients.extend(component.coefficients())
        # Define canonical numbering of arguments and coefficients
        self._arguments = tuple(sorted(set(arguments), key=lambda x: x.number()))
        self._coefficients = tuple(sorted(set(coefficients), key=lambda x: x.count()))

    def _analyze_domains(self):
        """Analyze which domains can be found in FormSum."""
        from ufl.domain import join_domains

        # Collect unique domains
        self._domains = join_domains(
            chain.from_iterable(e.ufl_domains() for e in self.ufl_operands)
        )

    def ufl_domains(self):
        """Return all domains found in the base form."""
        if self._domains is None:
            self._analyze_domains()
        return self._domains

    def __hash__(self):
        """Hash."""
        if self._hash is None:
            self._hash = hash(tuple(hash(component) for component in self.components()))
        return self._hash

    def equals(self, other):
        """Evaluate ``bool(lhs_form == rhs_form)``."""
        if type(other) is not FormSum:
            return False
        if self is other:
            return True
        return len(self.components()) == len(other.components()) and all(
            a == b for a, b in zip(self.components(), other.components())
        )

    def __str__(self):
        """Compute shorter string representation of form. This can be huge for complicated forms."""
        # Warning used for making sure we don't use this in the general pipeline:
        # warning("Calling str on form is potentially expensive and
        # should be avoided except during debugging.")
        # Not caching this because it can be huge
        s = "\n  +  ".join(str(component) for component in self.components())
        return s or "<empty FormSum>"

    def __repr__(self):
        """Compute repr string of form. This can be huge for complicated forms."""
        # Warning used for making sure we don't use this in the general pipeline:
        # warning("Calling repr on form is potentially expensive and
        # should be avoided except during debugging.")
        # Not caching this because it can be huge
        itgs = ", ".join(repr(component) for component in self.components())
        r = "FormSum([" + itgs + "])"
        return r


@ufl_type()
class ZeroBaseForm(BaseForm):
    """Description of a zero base form.

    ZeroBaseForm is idempotent with respect to assembly and is mostly
    used for sake of simplifying base-form expressions.
    """

    __slots__ = (
        "_arguments",
        "_coefficients",
        "ufl_operands",
        "_domains",
        "_hash",
        # Pyadjoint compatibility
        "form",
    )

    def __init__(self, arguments):
        """Initialise."""
        BaseForm.__init__(self)
        self._arguments = arguments
        self.ufl_operands = arguments
        self._hash = None
        self.form = None

    def _analyze_form_arguments(self):
        """Analyze form arguments."""
        # `self._arguments` is already set in `BaseForm.__init__`
        self._coefficients = ()

    def _analyze_domains(self):
        """Analyze which domains can be found in ZeroBaseForm."""
        from ufl.domain import join_domains

        # Collect unique domains
        self._domains = join_domains(
            chain.from_iterable(e.ufl_domains() for e in self.ufl_operands)
        )

    def ufl_domains(self):
        """Return all domains found in the base form."""
        if self._domains is None:
            self._analyze_domains()
        return self._domains

    def __ne__(self, other):
        """Overwrite BaseForm.__neq__ which relies on `equals`."""
        return not self == other

    def __eq__(self, other):
        """Check equality."""
        if type(other) is ZeroBaseForm:
            if self is other:
                return True
            return self._arguments == other._arguments
        elif isinstance(other, (int, float)):
            return other == 0
        else:
            return False

    def __str__(self):
        """Format as a string."""
        return "ZeroBaseForm(%s)" % (", ".join(str(arg) for arg in self._arguments))

    def __repr__(self):
        """Representation."""
        return "ZeroBaseForm(%s)" % (", ".join(repr(arg) for arg in self._arguments))

    def __hash__(self):
        """Hash."""
        if self._hash is None:
            self._hash = hash(("ZeroBaseForm", hash(self._arguments)))
        return self._hash



# Copyright (C) 2008 Anders Logg and Kristian B. Oelgaard
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# This simple example illustrates how forms can be defined on different sub domains.
# It is supported for all three integral types.
from ufl import FunctionSpace, Mesh, TestFunction, TrialFunction, dS, ds, dx, tetrahedron
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", tetrahedron, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
space = FunctionSpace(domain, element)

v = TestFunction(space)
u = TrialFunction(space)

a = (
    v * u * dx(0)
    + 10.0 * v * u * dx(1)
    + v * u * ds(0)
    + 2.0 * v * u * ds(1)
    + v("+") * u("+") * dS(0)
    + 4.3 * v("+") * u("+") * dS(1)
)



from ufl import (
    CellDiameter,
    CellVolume,
    Circumradius,
    FacetArea,
    FacetNormal,
    FunctionSpace,
    Index,
    Mesh,
    SpatialCoordinate,
    TestFunction,
    TrialFunction,
    as_matrix,
    as_ufl,
    as_vector,
    quadrilateral,
    tetrahedron,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_str_int_value(self):
    assert str(as_ufl(3)) == "3"


def test_str_float_value(self):
    assert str(as_ufl(3.14)) == "3.14"


def test_str_zero(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    assert str(as_ufl(0)) == "0"
    assert str(0 * x) == "0 (shape (2,))"
    assert str(0 * x * x[Index(42)]) == "0 (shape (2,), index labels (42,))"


def test_str_index(self):
    assert str(Index(3)) == "i_3"
    assert str(Index(42)) == "i_{42}"


def test_str_coordinate(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    assert str(SpatialCoordinate(domain)) == "x"
    assert str(SpatialCoordinate(domain)[0]) == "x[0]"


def test_str_normal(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    assert str(FacetNormal(domain)) == "n"
    assert str(FacetNormal(domain)[0]) == "n[0]"


def test_str_circumradius(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    assert str(Circumradius(domain)) == "circumradius"


def test_str_diameter(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    assert str(CellDiameter(domain)) == "diameter"


def test_str_facetarea(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    assert str(FacetArea(domain)) == "facetarea"


def test_str_volume(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    assert str(CellVolume(domain)) == "volume"


def test_str_scalar_argument(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v = TestFunction(
        FunctionSpace(domain, FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1))
    )
    u = TrialFunction(
        FunctionSpace(domain, FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1))
    )
    assert str(v) == "v_0"
    assert str(u) == "v_1"


# def test_str_vector_argument(self): # FIXME

# def test_str_scalar_coefficient(self): # FIXME

# def test_str_vector_coefficient(self): # FIXME


def test_str_list_vector():
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    x, y, z = SpatialCoordinate(domain)
    v = as_vector((x, y, z))
    assert str(v) == ("[%s, %s, %s]" % (x, y, z))


def test_str_list_vector_with_zero():
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    x, y, z = SpatialCoordinate(domain)
    v = as_vector((x, 0, 0))
    assert str(v) == ("[%s, 0, 0]" % (x,))


def test_str_list_matrix():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x, y = SpatialCoordinate(domain)
    v = as_matrix(((2 * x, 3 * y), (4 * x, 5 * y)))
    a = str(2 * x)
    b = str(3 * y)
    c = str(4 * x)
    d = str(5 * y)
    assert str(v) == ("[\n  [%s, %s],\n  [%s, %s]\n]" % (a, b, c, d))


def test_str_list_matrix_with_zero():
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    x, y = SpatialCoordinate(domain)
    v = as_matrix(((2 * x, 3 * y), (0, 0)))
    a = str(2 * x)
    b = str(3 * y)
    c = str(as_vector((0, 0)))
    assert str(v) == ("[\n  [%s, %s],\n%s\n]" % (a, b, c))


# FIXME: Add more tests for tensors collapsing
#        partly or completely into Zero!


def test_str_element():
    elem = FiniteElement("Q", quadrilateral, 1, (), identity_pullback, H1)
    assert (
        repr(elem)
        == 'ufl.finiteelement.FiniteElement("Q", quadrilateral, 1, (), IdentityPullback(), H1)'
    )
    assert str(elem) == "<Q1 on a quadrilateral>"



"""Utilites for sorting."""

# Copyright (C) 2008-2016 Johan Hake
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import warnings


def topological_sorting(nodes, edges):
    """Return a topologically sorted list of the nodes.

    Implemented algorithm from Wikipedia (http://en.wikipedia.org/wiki/Topological_sorting).

    No error for cyclic edges...
    """
    L = []
    S = nodes[:]
    for node in nodes:
        for es in edges.values():
            if node in es and node in S:
                S.remove(node)
                break

    while S:
        node = S.pop(0)
        L.append(node)
        node_edges = edges[node]
        while node_edges:
            m = node_edges.pop(0)
            found = False
            for es in edges.values():
                found = m in es
                if found:
                    break
            if not found:
                S.insert(0, m)

    return L


def sorted_by_count(seq):
    """Sort a sequence by the item.count()."""
    return sorted(seq, key=lambda x: x.count())


def sorted_by_key(mapping):
    """Sort dict items by key, allowing different key types."""

    # Python3 doesn't allow comparing builtins of different type,
    # therefore the typename trick here
    def _key(x):
        return (type(x[0]).__name__, x[0])

    return sorted(mapping.items(), key=_key)


def canonicalize_metadata(metadata):
    """Assuming metadata to be a dict with string keys and builtin python types as values.

    Transform dict to a tuple of (key, value) item tuples ordered by key,
    with dict, list and tuple values converted the same way recursively.
    Lists and tuples are converted to tuples. Other values are converted using str().
    This is such that the end result can be hashed and sorted using regular <,
    because python 3 doesn't allow e.g. (3 < "auto") which occurs regularly in metadata.
    """
    if metadata is None:
        return ()

    if isinstance(metadata, dict):
        keys = sorted(metadata.keys())
        assert all(isinstance(key, str) for key in keys)
        values = [metadata[key] for key in keys]
    elif isinstance(metadata, (tuple, list)):
        values = metadata

    newvalues = []
    for value in values:
        if isinstance(value, (dict, list, tuple)):
            value = canonicalize_metadata(value)
        elif isinstance(value, (int, float, str)) or value is None:
            value = str(value)
        elif hasattr(value, "ufl_signature"):
            value = value.ufl_signature
        else:
            warnings.warn(
                f"Applying str() to a metadata value of type {type(value).__name__}, "
                "don't know if this is safe."
            )
            value = str(value)
        newvalues.append(value)

    if isinstance(metadata, dict):
        return tuple(zip(keys, newvalues))
    else:
        return tuple(newvalues)



"""This module defines the Indexed class."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.constantvalue import Zero
from ufl.core.expr import Expr, ufl_err_str
from ufl.core.multiindex import FixedIndex, Index, MultiIndex
from ufl.core.operator import Operator
from ufl.core.ufl_type import ufl_type
from ufl.index_combination_utils import unique_sorted_indices
from ufl.precedence import parstr


@ufl_type(is_shaping=True, num_ops=2, is_terminal_modifier=True)
class Indexed(Operator):
    """Indexed expression."""

    __slots__ = (
        "ufl_free_indices",
        "ufl_index_dimensions",
    )

    def __new__(cls, expression, multiindex):
        """Create a new Indexed."""
        if isinstance(expression, Zero):
            # Zero-simplify indexed Zero objects
            shape = expression.ufl_shape
            efi = expression.ufl_free_indices
            efid = expression.ufl_index_dimensions
            fi = list(zip(efi, efid))
            for pos, ind in enumerate(multiindex._indices):
                if isinstance(ind, Index):
                    fi.append((ind.count(), shape[pos]))
            fi = unique_sorted_indices(sorted(fi))
            if fi:
                fi, fid = zip(*fi)
            else:
                fi, fid = (), ()
            return Zero(shape=(), free_indices=fi, index_dimensions=fid)
        elif expression.ufl_shape == () and multiindex == ():
            return expression
        else:
            return Operator.__new__(cls)

    def __init__(self, expression, multiindex):
        """Initialise."""
        # Store operands
        Operator.__init__(self, (expression, multiindex))

        # Error checking
        if not isinstance(expression, Expr):
            raise ValueError(f"Expecting Expr instance, not {ufl_err_str(expression)}.")
        if not isinstance(multiindex, MultiIndex):
            raise ValueError(f"Expecting MultiIndex instance, not {ufl_err_str(multiindex)}.")

        shape = expression.ufl_shape

        # Error checking
        if len(shape) != len(multiindex):
            raise ValueError(
                f"Invalid number of indices ({len(multiindex)}) for tensor "
                f"expression of rank {len(expression.ufl_shape)}:\n    {ufl_err_str(expression)}"
            )
        if any(
            int(di) >= int(si) or int(di) < 0
            for si, di in zip(shape, multiindex)
            if isinstance(di, FixedIndex)
        ):
            raise ValueError("Fixed index out of range!")

        # Build tuples of free index ids and dimensions
        efi = expression.ufl_free_indices
        efid = expression.ufl_index_dimensions
        fi = list(zip(efi, efid))
        for pos, ind in enumerate(multiindex._indices):
            if isinstance(ind, Index):
                fi.append((ind.count(), shape[pos]))
        fi = unique_sorted_indices(sorted(fi))
        if fi:
            fi, fid = zip(*fi)
        else:
            fi, fid = (), ()

        # Cache free index and dimensions
        self.ufl_free_indices = fi
        self.ufl_index_dimensions = fid

    ufl_shape = ()

    def evaluate(self, x, mapping, component, index_values, derivatives=()):
        """Evaluate."""
        A, ii = self.ufl_operands
        component = ii.evaluate(x, mapping, None, index_values)
        if derivatives:
            return A.evaluate(x, mapping, component, index_values, derivatives)
        else:
            return A.evaluate(x, mapping, component, index_values)

    def __str__(self):
        """Format as a string."""
        return "%s[%s]" % (parstr(self.ufl_operands[0], self), self.ufl_operands[1])

    def __getitem__(self, key):
        """Get an item."""
        if key == ():
            # So that one doesn't have to special case indexing of
            # expressions without shape.
            return self
        raise ValueError(
            f"Attempting to index with {ufl_err_str(key)}, "
            f"but object is already indexed: {ufl_err_str(self)}"
        )



# Copyright (C) 2008 Kristian B. Oelgaard
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2008-03-31
# Last changed: 2008-03-31
#
# The linearised bilinear form a(u,v) and linear form L(v) for
# the nonlinear equation - div (1+u) grad u = f (non-linear Poisson)
from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dot,
    dx,
    grad,
    i,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

QE = FiniteElement("Quadrature", triangle, 2, (), identity_pullback, H1)
sig = FiniteElement("Quadrature", triangle, 1, (2,), identity_pullback, H1)

qe_space = FunctionSpace(domain, QE)
sig_space = FunctionSpace(domain, sig)

v = TestFunction(space)
u = TrialFunction(space)
u0 = Coefficient(space)
C = Coefficient(qe_space)
sig0 = Coefficient(sig_space)
f = Coefficient(space)

a = (
    v.dx(i) * C * u.dx(i) * dx(metadata={"quadrature_degree": 2})
    + v.dx(i) * 2 * u0 * u * u0.dx(i) * dx
)
L = v * f * dx - dot(grad(v), sig0) * dx(metadata={"quadrature_degree": 1})



"""Test internal utility functions."""

from ufl.utils.indexflattening import flatten_multiindex, shape_to_strides, unflatten_index


def test_shape_to_strides():
    assert () == shape_to_strides(())
    assert (1,) == shape_to_strides((3,))
    assert (2, 1) == shape_to_strides((3, 2))
    assert (4, 1) == shape_to_strides((3, 4))
    assert (12, 4, 1) == shape_to_strides((6, 3, 4))


def test_flatten_multiindex_to_multiindex():
    sh = (2, 3, 5)
    strides = shape_to_strides(sh)
    for i in range(sh[2]):
        for j in range(sh[1]):
            for k in range(sh[0]):
                index = (k, j, i)
                c = flatten_multiindex(index, strides)
                index2 = unflatten_index(c, strides)
                assert index == index2


def test_indexing_to_component():
    assert 0 == flatten_multiindex((), shape_to_strides(()))
    assert 0 == flatten_multiindex((0,), shape_to_strides((2,)))
    assert 1 == flatten_multiindex((1,), shape_to_strides((2,)))
    assert 3 == flatten_multiindex((1, 1), shape_to_strides((2, 2)))
    for i in range(5):
        for j in range(3):
            for k in range(2):
                assert 15 * k + 5 * j + i == flatten_multiindex(
                    (k, j, i), shape_to_strides((2, 3, 5))
                )


def test_component_numbering():
    from ufl.permutation import build_component_numbering

    sh = (2, 2)
    sm = {(1, 0): (0, 1)}
    v, s = build_component_numbering(sh, sm)
    assert v == {(0, 1): 1, (1, 0): 1, (0, 0): 0, (1, 1): 2}
    assert s == [(0, 0), (0, 1), (1, 1)]

    sh = (3, 3)
    sm = {(1, 0): (0, 1), (2, 0): (0, 2), (2, 1): (1, 2)}
    v, s = build_component_numbering(sh, sm)
    assert v == {
        (0, 1): 1,
        (1, 2): 4,
        (0, 0): 0,
        (2, 1): 4,
        (1, 1): 3,
        (2, 0): 2,
        (2, 2): 5,
        (1, 0): 1,
        (0, 2): 2,
    }
    assert s == [(0, 0), (0, 1), (0, 2), (1, 1), (1, 2), (2, 2)]


def test_index_flattening():
    from ufl.utils.indexflattening import flatten_multiindex, shape_to_strides, unflatten_index

    # Scalar shape
    s = ()
    st = shape_to_strides(s)
    assert st == ()
    c = ()
    q = flatten_multiindex(c, st)
    c2 = unflatten_index(q, st)
    assert q == 0
    assert c2 == ()

    # Vector shape
    s = (2,)
    st = shape_to_strides(s)
    assert st == (1,)
    for i in range(s[0]):
        c = (i,)
        q = flatten_multiindex(c, st)
        c2 = unflatten_index(q, st)
        assert c == c2

    # Tensor shape
    s = (2, 3)
    st = shape_to_strides(s)
    assert st == (3, 1)
    for i in range(s[0]):
        for j in range(s[1]):
            c = (i, j)
            q = flatten_multiindex(c, st)
            c2 = unflatten_index(q, st)
            assert c == c2

    # Rank 3 tensor shape
    s = (2, 3, 4)
    st = shape_to_strides(s)
    assert st == (12, 4, 1)
    for i in range(s[0]):
        for j in range(s[1]):
            for k in range(s[2]):
                c = (i, j, k)
                q = flatten_multiindex(c, st)
                c2 = unflatten_index(q, st)
                assert c == c2

    # Taylor-Hood example:

    # pressure element is index 3:
    c = (3,)
    # get flat index:
    i = flatten_multiindex(c, shape_to_strides((4,)))
    # remove offset:
    i -= 3
    # map back to scalar component:
    c2 = unflatten_index(i, shape_to_strides(()))
    assert () == c2

    # vector element y-component is index 1:
    c = (1,)
    # get flat index:
    i = flatten_multiindex(c, shape_to_strides((4,)))
    # remove offset:
    i -= 0
    # map back to vector component:
    c2 = unflatten_index(i, shape_to_strides((3,)))
    assert (1,) == c2

    # Try a tensor/vector element:
    mixed_shape = (6,)
    ts = (2, 2)
    vs = (2,)
    offset = 4  # product(ts)

    # vector element y-component is index offset+1:
    c = (offset + 1,)
    # get flat index:
    i = flatten_multiindex(c, shape_to_strides(mixed_shape))
    # remove offset:
    i -= offset
    # map back to vector component:
    c2 = unflatten_index(i, shape_to_strides(vs))
    assert (1,) == c2

    for k in range(4):
        # tensor element (1,1)-component is index 3:
        c = (k,)
        # get flat index:
        i = flatten_multiindex(c, shape_to_strides(mixed_shape))
        # remove offset:
        i -= 0
        # map back to tensor component:
        c2 = unflatten_index(i, shape_to_strides(ts))
        assert (k // 2, k % 2) == c2


def test_stackdict():
    from ufl.utils.stacks import StackDict

    d = StackDict(a=1)
    assert d["a"] == 1
    d.push("a", 2)
    assert d["a"] == 2
    d.push("a", 3)
    d.push("b", 9)
    assert d["a"] == 3
    assert d["b"] == 9
    d.pop()
    assert d["a"] == 3
    assert "b" not in d
    d.pop()
    assert d["a"] == 2
    d.pop()
    assert d["a"] == 1



"""This module defines the single index types and some internal index utilities."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs and Anders Logg
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016.

from ufl.core.terminal import Terminal
from ufl.core.ufl_type import ufl_type
from ufl.utils.counted import Counted

# Export list for ufl.classes
__all_classes__ = ["IndexBase", "FixedIndex", "Index"]


class IndexBase(object):
    """Base class for all indices."""

    __slots__ = ()

    def __init__(self):
        """Initialise."""


class FixedIndex(IndexBase):
    """UFL value: An index with a specific value assigned."""

    __slots__ = ("_value", "_hash")

    _cache = {}

    def __getnewargs__(self):
        """Get new args."""
        return (self._value,)

    def __new__(cls, value):
        """Create new FixedIndex."""
        self = FixedIndex._cache.get(value)
        if self is None:
            if not isinstance(value, int):
                raise ValueError("Expecting integer value for fixed index.")
            self = IndexBase.__new__(cls)
            self._init(value)
            FixedIndex._cache[value] = self
        return self

    def _init(self, value):
        """Initialise."""
        IndexBase.__init__(self)
        self._value = value
        self._hash = hash(("FixedIndex", self._value))

    def __init__(self, value):
        """Initialise."""

    def __hash__(self):
        """Hash."""
        return self._hash

    def __eq__(self, other):
        """Check equality."""
        return isinstance(other, FixedIndex) and (self._value == other._value)

    def __int__(self):
        """Convert to int."""
        return self._value

    def __str__(self):
        """Represent with a string."""
        return f"{self._value}"

    def __repr__(self):
        """Return representation."""
        return f"FixedIndex({self._value})"


class Index(IndexBase, Counted):
    """UFL value: An index with no value assigned.

    Used to represent free indices in Einstein indexing notation.
    """

    __slots__ = ("_count", "_counted_class")

    def __init__(self, count=None):
        """Initialise."""
        IndexBase.__init__(self)
        Counted.__init__(self, count, Index)

    def __hash__(self):
        """Hash."""
        return hash(("Index", self._count))

    def __eq__(self, other):
        """Check equality."""
        return isinstance(other, Index) and (self._count == other._count)

    def __str__(self):
        """Represent as a string."""
        c = f"{self._count}"
        if len(c) > 1:
            c = f"{{{c}}}"
        return f"i_{c}"

    def __repr__(self):
        """Return representation."""
        return f"Index({self._count})"


@ufl_type()
class MultiIndex(Terminal):
    """Represents a sequence of indices, either fixed or free."""

    __slots__ = ("_indices",)

    _cache = {}

    def __getnewargs__(self):
        """Get new args."""
        return (self._indices,)

    def __new__(cls, indices):
        """Create new MultiIndex."""
        if not isinstance(indices, tuple):
            raise ValueError("Expecting a tuple of indices.")

        if all(isinstance(ind, FixedIndex) for ind in indices):
            # Cache multiindices consisting of purely fixed indices
            # (aka flyweight pattern)
            key = tuple(ind._value for ind in indices)
            self = MultiIndex._cache.get(key)
            if self is not None:
                return self
            self = Terminal.__new__(cls)
            MultiIndex._cache[key] = self
        else:
            # Create a new object if we have any free indices (too
            # many combinations to cache)
            if not all(isinstance(ind, IndexBase) for ind in indices):
                raise ValueError("Expecting only Index and FixedIndex objects.")
            self = Terminal.__new__(cls)

        # Initialize here instead of in __init__ to avoid overwriting
        # self._indices from cached objects
        self._init(indices)
        return self

    def __init__(self, indices):
        """Initialise."""

    def _init(self, indices):
        """Initialise."""
        Terminal.__init__(self)
        self._indices = indices

    def indices(self):
        """Return tuple of indices."""
        return self._indices

    def _ufl_compute_hash_(self):
        """Compute UFL hash."""
        return hash(("MultiIndex",) + tuple(hash(ind) for ind in self._indices))

    def __eq__(self, other):
        """Check equality."""
        return isinstance(other, MultiIndex) and self._indices == other._indices

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate index."""
        # Build component from index values
        component = []
        for i in self._indices:
            if isinstance(i, FixedIndex):
                component.append(i._value)
            elif isinstance(i, Index):
                component.append(index_values[i])
        return tuple(component)

    @property
    def ufl_shape(self):
        """Get the UFL shape.

        This should not be used.
        """
        raise ValueError("Multiindex has no shape (it is not a tensor expression).")

    @property
    def ufl_free_indices(self):
        """Get the UFL free indices.

        This should not be used.
        """
        raise ValueError("Multiindex has no free indices (it is not a tensor expression).")

    @property
    def ufl_index_dimensions(self):
        """Get the UFL index dimensions.

        This should not be used.
        """
        raise ValueError("Multiindex has no free indices (it is not a tensor expression).")

    def is_cellwise_constant(self):
        """Check if cellwise constant.

        Always True.
        """
        return True

    def ufl_domains(self):
        """Return tuple of domains related to this terminal object."""
        return ()

    # --- Adding multiindices ---

    def __add__(self, other):
        """Add."""
        if isinstance(other, tuple):
            return MultiIndex(self._indices + other)
        elif isinstance(other, MultiIndex):
            return MultiIndex(self._indices + other._indices)
        return NotImplemented

    def __radd__(self, other):
        """Add."""
        if isinstance(other, tuple):
            return MultiIndex(other + self._indices)
        elif isinstance(other, MultiIndex):
            return MultiIndex(other._indices + self._indices)
        return NotImplemented

    # --- String formatting ---

    def __str__(self):
        """Format as a string."""
        return ", ".join(str(i) for i in self._indices)

    def __repr__(self):
        """Return representation."""
        return f"MultiIndex({self._indices!r})"

    # --- Iteration protocol ---
    def __len__(self):
        """Get length."""
        return len(self._indices)

    def __getitem__(self, i):
        """Get an item."""
        return self._indices[i]

    def __iter__(self):
        """Return iteratable."""
        return iter(self._indices)


def indices(n):
    """Return a tuple of n new Index objects."""
    return tuple(Index() for i in range(n))



#
# Author: Anders Logg
# Modified by: Martin Sandve Alnes
# Date: 2009-01-12
#
from ufl import FunctionSpace, Mesh, TestFunction, TrialFunction, dx, grad, inner, tetrahedron
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
space = FunctionSpace(domain, element)

v = TestFunction(space)
u = TrialFunction(space)


def epsilon(v):
    Dv = grad(v)
    return 0.5 * (Dv + Dv.T)


a = inner(epsilon(v), epsilon(u)) * dx
